import logging
import os


def restore_build_artifacts_from_test_details(
    build_artifacts, conn, temporary_dir, testDetails
):
    for build_artifact in build_artifacts:
        build_artifact_key = "{}".format(build_artifact).encode()
        if build_artifact_key in testDetails:
            buffer_key = testDetails[build_artifact_key]
            logging.info(
                "Reading artifact binary {} from key {}".format(
                    build_artifact, buffer_key
                )
            )
            buffer = bytes(conn.get(buffer_key))
            artifact_fname = "{}/{}".format(temporary_dir, build_artifact)
            with open(artifact_fname, "wb") as fd:
                fd.write(buffer)
                os.chmod(artifact_fname, 755)

            logging.info(
                "Successfully restored {} into {}".format(
                    build_artifact, artifact_fname
                )
            )



import datetime
import json
import logging
import math
import os
import shutil
import sys
import tempfile
import traceback
from pathlib import Path
import re

import docker
import redis
from docker.models.containers import Container
from pytablewriter import CsvTableWriter, MarkdownTableWriter
from redisbench_admin.profilers.profilers_local import (
    check_compatible_system_and_kernel_and_prepare_profile,
    local_profilers_platform_checks,
    profilers_start_if_required,
    profilers_stop_if_required,
)
from redisbench_admin.run.common import (
    get_start_time_vars,
    merge_default_and_config_metrics,
    prepare_benchmark_parameters,
    dbconfig_keyspacelen_check,
)
from redisbench_admin.run.metrics import extract_results_table
from redisbench_admin.run.run import calculate_client_tool_duration_and_check
from redisbench_admin.utils.benchmark_config import (
    get_final_benchmark_config,
    get_defaults,
)
from redisbench_admin.utils.local import get_local_run_full_filename
from redisbench_admin.utils.results import post_process_benchmark_results

from redis_benchmarks_specification.__common__.env import (
    LOG_DATEFMT,
    LOG_FORMAT,
    LOG_LEVEL,
    REDIS_HEALTH_CHECK_INTERVAL,
    REDIS_SOCKET_TIMEOUT,
    S3_BUCKET_NAME,
)
from redis_benchmarks_specification.__common__.package import (
    get_version_string,
    populate_with_poetry_data,
)
from redis_benchmarks_specification.__common__.runner import (
    extract_testsuites,
    exporter_datasink_common,
    reset_commandstats,
    execute_init_commands,
)
from redis_benchmarks_specification.__common__.spec import (
    extract_client_container_image,
    extract_client_cpu_limit,
    extract_client_tool,
)
from redis_benchmarks_specification.__runner__.args import create_client_runner_args


def parse_size(size):
    units = {
        "B": 1,
        "KB": 2**10,
        "MB": 2**20,
        "GB": 2**30,
        "TB": 2**40,
        "": 1,
        "KIB": 10**3,
        "MIB": 10**6,
        "GIB": 10**9,
        "TIB": 10**12,
        "K": 2**10,
        "M": 2**20,
        "G": 2**30,
        "T": 2**40,
        "": 1,
        "KI": 10**3,
        "MI": 10**6,
        "GI": 10**9,
        "TI": 10**12,
    }
    m = re.match(r"^([\d\.]+)\s*([a-zA-Z]{0,3})$", str(size).strip())
    number, unit = float(m.group(1)), m.group(2).upper()
    return int(number * units[unit])


def main():
    _, _, project_version = populate_with_poetry_data()
    project_name_suffix = "redis-benchmarks-spec-client-runner"
    project_name = f"{project_name_suffix} (solely client)"
    parser = create_client_runner_args(
        get_version_string(project_name, project_version)
    )
    args = parser.parse_args()

    run_client_runner_logic(args, project_name, project_name_suffix, project_version)


def run_client_runner_logic(args, project_name, project_name_suffix, project_version):
    if args.logname is not None:
        print(f"Writting log to {args.logname}")
        logging.basicConfig(
            filename=args.logname,
            filemode="a",
            format=LOG_FORMAT,
            datefmt=LOG_DATEFMT,
            level=LOG_LEVEL,
        )
    else:
        # logging settings
        logging.basicConfig(
            format=LOG_FORMAT,
            level=LOG_LEVEL,
            datefmt=LOG_DATEFMT,
        )
    logging.info(get_version_string(project_name, project_version))
    testsuite_spec_files = extract_testsuites(args)
    datasink_conn = None
    if args.datasink_push_results_redistimeseries:
        logging.info(
            "Checking redistimeseries datasink connection is available at: {}:{} to push the timeseries data".format(
                args.datasink_redistimeseries_host, args.datasink_redistimeseries_port
            )
        )
        try:
            datasink_conn = redis.StrictRedis(
                host=args.datasink_redistimeseries_host,
                port=args.datasink_redistimeseries_port,
                decode_responses=True,
                password=args.datasink_redistimeseries_pass,
                username=args.datasink_redistimeseries_user,
                health_check_interval=REDIS_HEALTH_CHECK_INTERVAL,
                socket_connect_timeout=REDIS_SOCKET_TIMEOUT,
                socket_keepalive=True,
            )
            datasink_conn.ping()
            datasink_conn.client_setname(project_name_suffix)
        except redis.exceptions.ConnectionError as e:
            logging.error(
                "Unable to connect to redis available at: {}:{}".format(
                    args.datasink_redistimeseries_host,
                    args.datasink_redistimeseries_port,
                )
            )
            logging.error(f"Error message {e.__str__()}")
            exit(1)
    running_platform = args.platform_name
    tls_enabled = args.tls
    tls_skip_verify = args.tls_skip_verify
    tls_cert = args.cert
    tls_key = args.key
    tls_cacert = args.cacert
    resp_version = args.resp
    client_aggregated_results_folder = args.client_aggregated_results_folder
    preserve_temporary_client_dirs = args.preserve_temporary_client_dirs

    docker_client = docker.from_env()
    home = str(Path.home())
    profilers_list = []
    profilers_enabled = args.enable_profilers
    if profilers_enabled:
        profilers_list = args.profilers.split(",")
        res = check_compatible_system_and_kernel_and_prepare_profile(args)
        if res is False:
            logging.error(
                "Requested for the following profilers to be enabled but something went wrong: {}.".format(
                    " ".join(profilers_list)
                )
            )
            exit(1)
    override_memtier_test_time = args.override_memtier_test_time
    if override_memtier_test_time > 0:
        logging.info(
            "Overriding memtier benchmark --test-time to {} seconds".format(
                override_memtier_test_time
            )
        )
    logging.info("Running the benchmark specs.")
    process_self_contained_coordinator_stream(
        args,
        args.datasink_push_results_redistimeseries,
        docker_client,
        home,
        None,
        datasink_conn,
        testsuite_spec_files,
        {},
        running_platform,
        profilers_enabled,
        profilers_list,
        tls_enabled,
        tls_skip_verify,
        tls_cert,
        tls_key,
        tls_cacert,
        client_aggregated_results_folder,
        preserve_temporary_client_dirs,
        resp_version,
        override_memtier_test_time,
    )


def prepare_memtier_benchmark_parameters(
    clientconfig,
    full_benchmark_path,
    port,
    server,
    password,
    local_benchmark_output_filename,
    oss_cluster_api_enabled=False,
    tls_enabled=False,
    tls_skip_verify=False,
    tls_cert=None,
    tls_key=None,
    tls_cacert=None,
    resp_version=None,
    override_memtier_test_time=0,
    override_test_runs=1,
):
    arbitrary_command = False
    benchmark_command = [
        full_benchmark_path,
        "--port",
        f"{port}",
        "--server",
        f"{server}",
        "--json-out-file",
        local_benchmark_output_filename,
    ]
    if password is not None:
        benchmark_command.extend(["--authenticate", password])
    if tls_enabled:
        benchmark_command.append("--tls")
        if tls_cert is not None and tls_cert != "":
            benchmark_command.extend(["--cert", tls_cert])
        if tls_key is not None and tls_key != "":
            benchmark_command.extend(["--key", tls_key])
        if tls_cacert is not None and tls_cacert != "":
            benchmark_command.extend(["--cacert", tls_cacert])
        if tls_skip_verify:
            benchmark_command.append("--tls-skip-verify")

    if resp_version:
        tool = clientconfig["tool"]
        if tool == "memtier_benchmark":
            if resp_version == "3":
                benchmark_command.extend(["--protocol", "resp{}".format(resp_version)])
        elif tool == "redis-benchmark":
            if resp_version == "3":
                benchmark_command.append("-3")

    if oss_cluster_api_enabled is True:
        benchmark_command.append("--cluster-mode")
    logging.info(f"Preparing the benchmark parameters. {benchmark_command}.")
    benchmark_command_str = " ".join(benchmark_command)
    if "arguments" in clientconfig:
        benchmark_command_str = benchmark_command_str + " " + clientconfig["arguments"]

    if "--command" in benchmark_command_str:
        arbitrary_command = True

    if override_test_runs > 1:
        benchmark_command_str = re.sub(
            "--run-count\\s\\d+",
            "--run-count={}".format(override_test_runs),
            benchmark_command_str,
        )
        benchmark_command_str = re.sub(
            "--run-count=\\d+",
            "--run-count={}".format(override_test_runs),
            benchmark_command_str,
        )
        benchmark_command_str = re.sub(
            '--run-count="\\d+"',
            "--run-count={}".format(override_test_runs),
            benchmark_command_str,
        )
        # short
        benchmark_command_str = re.sub(
            "-x\\s\\d+",
            "-x={}".format(override_test_runs),
            benchmark_command_str,
        )
        benchmark_command_str = re.sub(
            "-x=\\d+",
            "-x={}".format(override_test_runs),
            benchmark_command_str,
        )
        benchmark_command_str = re.sub(
            '-x="\\d+"',
            "-x={}".format(override_test_runs),
            benchmark_command_str,
        )
        if (
            len(
                re.findall(
                    "--run-count={}".format(override_test_runs),
                    benchmark_command_str,
                )
            )
            == 0
            and len(
                re.findall(
                    "-x={}".format(override_test_runs),
                    benchmark_command_str,
                )
            )
            == 0
        ):
            logging.info("adding --run-count option to benchmark run. ")
            benchmark_command_str = (
                benchmark_command_str
                + " "
                + "--run-count={}".format(override_test_runs)
            )

    if override_memtier_test_time > 0:
        benchmark_command_str = re.sub(
            "--test-time\\s\\d+",
            "--test-time={}".format(override_memtier_test_time),
            benchmark_command_str,
        )
        benchmark_command_str = re.sub(
            "--test-time=\\d+",
            "--test-time={}".format(override_memtier_test_time),
            benchmark_command_str,
        )
        benchmark_command_str = re.sub(
            '--test-time="\\d+"',
            "--test-time={}".format(override_memtier_test_time),
            benchmark_command_str,
        )

    return None, benchmark_command_str, arbitrary_command


def process_self_contained_coordinator_stream(
    args,
    datasink_push_results_redistimeseries,
    docker_client,
    home,
    newTestInfo,
    datasink_conn,
    testsuite_spec_files,
    topologies_map,
    running_platform,
    profilers_enabled=False,
    profilers_list=[],
    tls_enabled=False,
    tls_skip_verify=False,
    tls_cert=None,
    tls_key=None,
    tls_cacert=None,
    client_aggregated_results_folder="",
    preserve_temporary_client_dirs=False,
    resp_version=None,
    override_memtier_test_time=0,
    used_memory_check_fail=False,
):
    def delete_temporary_files(
        temporary_dir_client, full_result_path, benchmark_tool_global
    ):
        if preserve_temporary_client_dirs is True:
            logging.info(f"Preserving temporary client dir {temporary_dir_client}")
        else:
            if "redis-benchmark" in benchmark_tool_global:
                if full_result_path is not None:
                    os.remove(full_result_path)
                    logging.info("Removing temporary JSON file")
            shutil.rmtree(temporary_dir_client, ignore_errors=True)
            logging.info(f"Removing temporary client dir {temporary_dir_client}")

    overall_result = True
    results_matrix = []
    total_test_suite_runs = 0
    dry_run_count = 0
    dry_run = args.dry_run
    dry_run_include_preload = args.dry_run_include_preload
    defaults_filename = args.defaults_filename
    override_test_runs = args.override_test_runs
    (
        _,
        _,
        default_metrics,
        _,
        _,
        _,
    ) = get_defaults(defaults_filename)

    for test_file in testsuite_spec_files:
        if defaults_filename in test_file:
            continue
        client_containers = []

        with open(test_file, "r") as stream:
            _, benchmark_config, test_name = get_final_benchmark_config(
                None, None, stream, ""
            )

            if tls_enabled:
                test_name = test_name + "-tls"
                logging.info(
                    "Given that TLS is enabled, appending -tls to the testname: {}.".format(
                        test_name
                    )
                )

            for topology_spec_name in benchmark_config["redis-topologies"]:
                test_result = False
                benchmark_tool_global = ""
                full_result_path = None
                try:
                    current_cpu_pos = args.cpuset_start_pos
                    temporary_dir_client = tempfile.mkdtemp(dir=home)

                    tf_github_org = args.github_org
                    tf_github_repo = args.github_repo
                    tf_triggering_env = args.platform_name
                    setup_type = args.setup_type
                    priority_upper_limit = args.tests_priority_upper_limit
                    priority_lower_limit = args.tests_priority_lower_limit
                    git_hash = "NA"
                    git_version = args.github_version
                    build_variant_name = "NA"
                    git_branch = None

                    port = args.db_server_port
                    host = args.db_server_host
                    password = args.db_server_password
                    oss_cluster_api_enabled = args.cluster_mode
                    ssl_cert_reqs = "required"
                    if tls_skip_verify:
                        ssl_cert_reqs = None
                    r = redis.StrictRedis(
                        host=host,
                        port=port,
                        password=password,
                        ssl=tls_enabled,
                        ssl_cert_reqs=ssl_cert_reqs,
                        ssl_keyfile=tls_key,
                        ssl_certfile=tls_cert,
                        ssl_ca_certs=tls_cacert,
                        ssl_check_hostname=False,
                    )
                    setup_name = "oss-standalone"
                    r.ping()
                    redis_conns = [r]
                    if oss_cluster_api_enabled:
                        redis_conns = []
                        logging.info("updating redis connections from cluster slots")
                        slots = r.cluster("slots")
                        for slot in slots:
                            # Master for slot range represented as nested networking information starts at pos 2
                            # example: [0, 5460, [b'127.0.0.1', 30001, b'eccd21c2e7e9b7820434080d2e394cb8f2a7eff2', []]]
                            slot_network_info = slot[2]
                            prefered_endpoint = slot_network_info[0]
                            prefered_port = slot_network_info[1]
                            shard_conn = redis.StrictRedis(
                                host=prefered_endpoint,
                                port=prefered_port,
                                password=password,
                                ssl=tls_enabled,
                                ssl_cert_reqs=ssl_cert_reqs,
                                ssl_keyfile=tls_key,
                                ssl_certfile=tls_cert,
                                ssl_ca_certs=tls_cacert,
                                ssl_check_hostname=False,
                            )
                            redis_conns.append(shard_conn)
                        logging.info(
                            "There are a total of {} shards".format(len(redis_conns))
                        )
                        setup_name = "oss-cluster"

                    redis_pids = []
                    for conn in redis_conns:
                        redis_pid = conn.info()["process_id"]
                        redis_pids.append(redis_pid)

                    github_actor = f"{tf_triggering_env}-{running_platform}"
                    dso = "redis-server"
                    profilers_artifacts_matrix = []

                    collection_summary_str = ""
                    if profilers_enabled:
                        collection_summary_str = local_profilers_platform_checks(
                            dso,
                            github_actor,
                            git_branch,
                            tf_github_repo,
                            git_hash,
                        )
                        logging.info(
                            "Using the following collection summary string for profiler description: {}".format(
                                collection_summary_str
                            )
                        )

                    ceil_client_cpu_limit = extract_client_cpu_limit(benchmark_config)
                    client_cpuset_cpus, current_cpu_pos = generate_cpuset_cpus(
                        ceil_client_cpu_limit, current_cpu_pos
                    )
                    if args.flushall_on_every_test_start:
                        logging.info("Sending FLUSHALL to the DB")
                        for conn in redis_conns:
                            conn.flushall()

                    benchmark_required_memory = get_benchmark_required_memory(
                        benchmark_config
                    )
                    maxmemory = 0
                    if args.maxmemory > 0:
                        maxmemory = args.maxmemory
                    else:
                        for conn in redis_conns:
                            maxmemory = maxmemory + get_maxmemory(conn)
                    if benchmark_required_memory > maxmemory:
                        logging.warning(
                            "Skipping test {} given maxmemory of server is bellow the benchmark required memory: {} < {}".format(
                                test_name, maxmemory, benchmark_required_memory
                            )
                        )
                        delete_temporary_files(
                            temporary_dir_client=temporary_dir_client,
                            full_result_path=None,
                            benchmark_tool_global=benchmark_tool_global,
                        )
                        continue

                    reset_commandstats(redis_conns)

                    client_mnt_point = "/mnt/client/"
                    benchmark_tool_workdir = client_mnt_point

                    metadata = {}
                    test_tls_cacert = None
                    test_tls_cert = None
                    test_tls_key = None
                    if tls_enabled:
                        metadata["tls"] = "true"
                        if tls_cert is not None and tls_cert != "":
                            _, test_tls_cert = cp_to_workdir(
                                temporary_dir_client, tls_cert
                            )
                        if tls_cacert is not None and tls_cacert != "":
                            _, test_tls_cacert = cp_to_workdir(
                                temporary_dir_client, tls_cacert
                            )
                        if tls_key is not None and tls_key != "":
                            _, test_tls_key = cp_to_workdir(
                                temporary_dir_client, tls_key
                            )
                    priority = None
                    if "priority" in benchmark_config:
                        priority = benchmark_config["priority"]

                        if priority is not None:
                            if priority > priority_upper_limit:
                                logging.warning(
                                    "Skipping test {} giving the priority limit ({}) is above the priority value ({})".format(
                                        test_name, priority_upper_limit, priority
                                    )
                                )
                                delete_temporary_files(
                                    temporary_dir_client=temporary_dir_client,
                                    full_result_path=None,
                                    benchmark_tool_global=benchmark_tool_global,
                                )
                                continue
                            if priority < priority_lower_limit:
                                logging.warning(
                                    "Skipping test {} giving the priority limit ({}) is bellow the priority value ({})".format(
                                        test_name, priority_lower_limit, priority
                                    )
                                )
                                delete_temporary_files(
                                    temporary_dir_client=temporary_dir_client,
                                    full_result_path=None,
                                    benchmark_tool_global=benchmark_tool_global,
                                )
                                continue
                            logging.info(
                                "Test {} priority ({}) is within the priority limit [{},{}]".format(
                                    test_name,
                                    priority,
                                    priority_lower_limit,
                                    priority_upper_limit,
                                )
                            )
                    if "dbconfig" in benchmark_config:
                        if "dataset" in benchmark_config["dbconfig"]:
                            if args.run_tests_with_dataset is False:
                                logging.warning(
                                    "Skipping test {} giving it implies dataset preload".format(
                                        test_name
                                    )
                                )
                                delete_temporary_files(
                                    temporary_dir_client=temporary_dir_client,
                                    full_result_path=None,
                                    benchmark_tool_global=benchmark_tool_global,
                                )
                                continue

                    if dry_run is True:
                        dry_run_count = dry_run_count + 1
                        delete_temporary_files(
                            temporary_dir_client=temporary_dir_client,
                            full_result_path=None,
                            benchmark_tool_global=benchmark_tool_global,
                        )
                        continue
                    if "dbconfig" in benchmark_config:
                        if "preload_tool" in benchmark_config["dbconfig"]:
                            res = data_prepopulation_step(
                                benchmark_config,
                                benchmark_tool_workdir,
                                client_cpuset_cpus,
                                docker_client,
                                git_hash,
                                port,
                                temporary_dir_client,
                                test_name,
                                host,
                                tls_enabled,
                                tls_skip_verify,
                                test_tls_cert,
                                test_tls_key,
                                test_tls_cacert,
                                resp_version,
                                args.benchmark_local_install,
                                password,
                                oss_cluster_api_enabled,
                            )
                            if res is False:
                                logging.warning(
                                    "Skipping this test given preload result was false"
                                )
                                delete_temporary_files(
                                    temporary_dir_client=temporary_dir_client,
                                    full_result_path=None,
                                    benchmark_tool_global=benchmark_tool_global,
                                )
                                continue
                    execute_init_commands(
                        benchmark_config, r, dbconfig_keyname="dbconfig"
                    )

                    used_memory_check(
                        test_name,
                        benchmark_required_memory,
                        redis_conns,
                        "start of benchmark",
                        used_memory_check_fail,
                    )

                    logging.info("Checking if there is a keyspace check being enforced")
                    dbconfig_keyspacelen_check(
                        benchmark_config,
                        redis_conns,
                    )

                    if dry_run_include_preload is True:
                        dry_run_count = dry_run_count + 1
                        delete_temporary_files(
                            temporary_dir_client=temporary_dir_client,
                            full_result_path=None,
                            benchmark_tool_global=benchmark_tool_global,
                        )
                        continue

                    benchmark_tool = extract_client_tool(benchmark_config)
                    benchmark_tool_global = benchmark_tool
                    # backwards compatible
                    if benchmark_tool is None:
                        benchmark_tool = "redis-benchmark"
                    full_benchmark_path = f"/usr/local/bin/{benchmark_tool}"

                    # setup the benchmark
                    (
                        start_time,
                        start_time_ms,
                        start_time_str,
                    ) = get_start_time_vars()
                    local_benchmark_output_filename = get_local_run_full_filename(
                        start_time_str,
                        git_hash,
                        test_name,
                        setup_type,
                    )
                    logging.info(
                        "Will store benchmark json output to local file {}".format(
                            local_benchmark_output_filename
                        )
                    )
                    arbitrary_command = False

                    if "memtier_benchmark" not in benchmark_tool:
                        # prepare the benchmark command
                        (
                            benchmark_command,
                            benchmark_command_str,
                        ) = prepare_benchmark_parameters(
                            benchmark_config,
                            full_benchmark_path,
                            port,
                            host,
                            local_benchmark_output_filename,
                            False,
                            benchmark_tool_workdir,
                            False,
                        )
                    else:
                        (
                            _,
                            benchmark_command_str,
                            arbitrary_command,
                        ) = prepare_memtier_benchmark_parameters(
                            benchmark_config["clientconfig"],
                            full_benchmark_path,
                            port,
                            host,
                            password,
                            local_benchmark_output_filename,
                            oss_cluster_api_enabled,
                            tls_enabled,
                            tls_skip_verify,
                            test_tls_cert,
                            test_tls_key,
                            test_tls_cacert,
                            resp_version,
                            override_memtier_test_time,
                            override_test_runs,
                        )

                    if (
                        arbitrary_command
                        and oss_cluster_api_enabled
                        and "memtier" in benchmark_tool
                    ):
                        logging.warning(
                            "Forcing skip this test given there is an arbitrary commmand and memtier usage. Check https://github.com/RedisLabs/memtier_benchmark/pull/117 ."
                        )
                        delete_temporary_files(
                            temporary_dir_client=temporary_dir_client,
                            full_result_path=None,
                            benchmark_tool_global=benchmark_tool_global,
                        )
                        continue

                    client_container_image = extract_client_container_image(
                        benchmark_config
                    )
                    profiler_call_graph_mode = "dwarf"
                    profiler_frequency = 99

                    # start the profile
                    (
                        profiler_name,
                        profilers_map,
                    ) = profilers_start_if_required(
                        profilers_enabled,
                        profilers_list,
                        redis_pids,
                        setup_name,
                        start_time_str,
                        test_name,
                        profiler_frequency,
                        profiler_call_graph_mode,
                    )

                    # run the benchmark
                    benchmark_start_time = datetime.datetime.now()

                    if args.benchmark_local_install:
                        logging.info("Running memtier benchmark outside of docker")
                        benchmark_command_str = (
                            "taskset -c "
                            + client_cpuset_cpus
                            + " "
                            + benchmark_command_str
                        )
                        logging.info(
                            "Running memtier benchmark command {}".format(
                                benchmark_command_str
                            )
                        )
                        stream = os.popen(benchmark_command_str)
                        client_container_stdout = stream.read()
                        move_command = "mv {} {}".format(
                            local_benchmark_output_filename, temporary_dir_client
                        )
                        os.system(move_command)
                    else:
                        logging.info(
                            "Using docker image {} as benchmark client image (cpuset={}) with the following args: {}".format(
                                client_container_image,
                                client_cpuset_cpus,
                                benchmark_command_str,
                            )
                        )

                        client_container_stdout = docker_client.containers.run(
                            image=client_container_image,
                            volumes={
                                temporary_dir_client: {
                                    "bind": client_mnt_point,
                                    "mode": "rw",
                                },
                            },
                            auto_remove=True,
                            privileged=True,
                            working_dir=benchmark_tool_workdir,
                            command=benchmark_command_str,
                            network_mode="host",
                            detach=False,
                            cpuset_cpus=client_cpuset_cpus,
                        )

                    benchmark_end_time = datetime.datetime.now()
                    benchmark_duration_seconds = (
                        calculate_client_tool_duration_and_check(
                            benchmark_end_time, benchmark_start_time
                        )
                    )
                    (
                        _,
                        overall_tabular_data_map,
                    ) = profilers_stop_if_required(
                        datasink_push_results_redistimeseries,
                        benchmark_duration_seconds,
                        collection_summary_str,
                        dso,
                        tf_github_org,
                        tf_github_repo,
                        profiler_name,
                        profilers_artifacts_matrix,
                        profilers_enabled,
                        profilers_map,
                        redis_pids,
                        S3_BUCKET_NAME,
                        test_name,
                    )

                    logging.info("Printing client tool stdout output")

                    used_memory_check(
                        test_name,
                        benchmark_required_memory,
                        redis_conns,
                        "end of benchmark",
                        used_memory_check_fail,
                    )

                    if args.flushall_on_every_test_end:
                        logging.info("Sending FLUSHALL to the DB")
                        for r in redis_conns:
                            r.flushall()
                    datapoint_time_ms = start_time_ms

                    post_process_benchmark_results(
                        benchmark_tool,
                        local_benchmark_output_filename,
                        datapoint_time_ms,
                        start_time_str,
                        client_container_stdout,
                        None,
                    )
                    full_result_path = local_benchmark_output_filename
                    if "memtier_benchmark" in benchmark_tool:
                        full_result_path = "{}/{}".format(
                            temporary_dir_client, local_benchmark_output_filename
                        )
                    logging.info(f"Reading results json from {full_result_path}")

                    with open(
                        full_result_path,
                        "r",
                    ) as json_file:
                        results_dict = json.load(json_file)
                        print_results_table_stdout(
                            benchmark_config,
                            default_metrics,
                            results_dict,
                            setup_type,
                            test_name,
                            None,
                        )
                        prepare_overall_total_test_results(
                            benchmark_config,
                            default_metrics,
                            results_dict,
                            test_name,
                            results_matrix,
                        )

                    dataset_load_duration_seconds = 0

                    exporter_datasink_common(
                        benchmark_config,
                        benchmark_duration_seconds,
                        build_variant_name,
                        datapoint_time_ms,
                        dataset_load_duration_seconds,
                        datasink_conn,
                        datasink_push_results_redistimeseries,
                        git_branch,
                        git_version,
                        metadata,
                        redis_conns,
                        results_dict,
                        running_platform,
                        setup_name,
                        setup_type,
                        test_name,
                        tf_github_org,
                        tf_github_repo,
                        tf_triggering_env,
                        topology_spec_name,
                        default_metrics,
                    )
                    test_result = True
                    total_test_suite_runs = total_test_suite_runs + 1

                except:
                    logging.critical(
                        "Some unexpected exception was caught "
                        "during local work. Failing test...."
                    )
                    logging.critical(sys.exc_info()[0])
                    print("-" * 60)
                    traceback.print_exc(file=sys.stdout)
                    print("-" * 60)
                    test_result = False
                # tear-down
                logging.info("Tearing down setup")
                for container in client_containers:
                    if type(container) == Container:
                        try:
                            container.stop()
                        except docker.errors.NotFound:
                            logging.info(
                                "When trying to stop Client container with id {} and image {} it was already stopped".format(
                                    container.id, container.image
                                )
                            )
                            pass

                if client_aggregated_results_folder != "":
                    os.makedirs(client_aggregated_results_folder, exist_ok=True)
                    dest_fpath = "{}/{}".format(
                        client_aggregated_results_folder,
                        local_benchmark_output_filename,
                    )
                    logging.info(
                        "Preserving local results file {} into {}".format(
                            full_result_path, dest_fpath
                        )
                    )
                    shutil.copy(full_result_path, dest_fpath)
                overall_result &= test_result

                delete_temporary_files(
                    temporary_dir_client=temporary_dir_client,
                    full_result_path=full_result_path,
                    benchmark_tool_global=benchmark_tool_global,
                )

    table_name = "Results for entire test-suite"
    results_matrix_headers = [
        "Test Name",
        "Metric JSON Path",
        "Metric Value",
    ]
    writer = MarkdownTableWriter(
        table_name=table_name,
        headers=results_matrix_headers,
        value_matrix=results_matrix,
    )
    writer.write_table()

    if client_aggregated_results_folder != "":
        os.makedirs(client_aggregated_results_folder, exist_ok=True)
        dest_fpath = f"{client_aggregated_results_folder}/aggregate-results.csv"
        logging.info(f"Storing an aggregated results CSV into {full_result_path}")

        csv_writer = CsvTableWriter(
            table_name=table_name,
            headers=results_matrix_headers,
            value_matrix=results_matrix,
        )
        csv_writer.dump(dest_fpath)

    if dry_run is True:
        logging.info(
            "Number of tests that would have been run: {}".format(dry_run_count)
        )


def get_maxmemory(r):
    maxmemory = int(r.info("memory")["maxmemory"])
    if maxmemory == 0:
        total_system_memory = int(r.info("memory")["total_system_memory"])
        logging.info(" Using total system memory as max {}".format(total_system_memory))
        maxmemory = total_system_memory
    else:
        logging.info(" Detected redis maxmemory config value {}".format(maxmemory))

    return maxmemory


def get_benchmark_required_memory(benchmark_config):
    benchmark_required_memory = 0
    if "dbconfig" in benchmark_config:
        if "resources" in benchmark_config["dbconfig"]:
            resources = benchmark_config["dbconfig"]["resources"]
            if "requests" in resources:
                resources_requests = benchmark_config["dbconfig"]["resources"][
                    "requests"
                ]
                if "memory" in resources_requests:
                    benchmark_required_memory = resources_requests["memory"]
                    benchmark_required_memory = int(
                        parse_size(benchmark_required_memory)
                    )
                    logging.info(
                        "Benchmark required memory: {} Bytes".format(
                            benchmark_required_memory
                        )
                    )
    return benchmark_required_memory


def used_memory_check(
    test_name,
    benchmark_required_memory,
    redis_conns,
    stage,
    used_memory_check_fail=False,
):
    used_memory = 0
    for conn in redis_conns:
        used_memory = used_memory + conn.info("memory")["used_memory"]
    used_memory_gb = int(math.ceil(float(used_memory) / 1024.0 / 1024.0 / 1024.0))
    logging.info("Benchmark used memory at {}: {}g".format(stage, used_memory_gb))
    if used_memory > benchmark_required_memory:
        logging.error(
            "The benchmark {} specified a dbconfig resource request of memory ({}) bellow the REAL MEMORY USAGE OF: {}. FIX IT!.".format(
                test_name, benchmark_required_memory, used_memory_gb
            )
        )
        if used_memory_check_fail:
            exit(1)


def cp_to_workdir(benchmark_tool_workdir, srcfile):
    head, filename = os.path.split(srcfile)
    dstfile = f"{benchmark_tool_workdir}/{filename}"
    shutil.copyfile(srcfile, dstfile)
    logging.info(
        f"Copying to workdir the following file {srcfile}. Final workdir file {dstfile}"
    )
    return dstfile, filename


def print_results_table_stdout(
    benchmark_config,
    default_metrics,
    results_dict,
    setup_name,
    test_name,
    cpu_usage=None,
):
    # check which metrics to extract
    (
        _,
        metrics,
    ) = merge_default_and_config_metrics(
        benchmark_config,
        default_metrics,
        None,
    )
    table_name = f"Results for {test_name} test-case on {setup_name} topology"
    results_matrix_headers = [
        "Metric JSON Path",
        "Metric Value",
    ]
    results_matrix = extract_results_table(metrics, results_dict)

    results_matrix = [[x[0], f"{x[3]:.3f}"] for x in results_matrix]
    writer = MarkdownTableWriter(
        table_name=table_name,
        headers=results_matrix_headers,
        value_matrix=results_matrix,
    )
    writer.write_table()


def prepare_overall_total_test_results(
    benchmark_config, default_metrics, results_dict, test_name, overall_results_matrix
):
    # check which metrics to extract
    (
        _,
        metrics,
    ) = merge_default_and_config_metrics(
        benchmark_config,
        default_metrics,
        None,
    )
    current_test_results_matrix = extract_results_table(metrics, results_dict)
    current_test_results_matrix = [
        [test_name, x[0], f"{x[3]:.3f}"] for x in current_test_results_matrix
    ]
    overall_results_matrix.extend(current_test_results_matrix)


def data_prepopulation_step(
    benchmark_config,
    benchmark_tool_workdir,
    client_cpuset_cpus,
    docker_client,
    git_hash,
    port,
    temporary_dir,
    test_name,
    host,
    tls_enabled=False,
    tls_skip_verify=False,
    tls_cert=None,
    tls_key=None,
    tls_cacert=None,
    resp_version=None,
    benchmark_local_install=False,
    password=None,
    oss_cluster_api_enabled=False,
):
    result = True
    # setup the benchmark
    (
        start_time,
        start_time_ms,
        start_time_str,
    ) = get_start_time_vars()
    local_benchmark_output_filename = get_local_run_full_filename(
        start_time_str,
        git_hash,
        "preload__" + test_name,
        "oss-standalone",
    )
    preload_image = extract_client_container_image(
        benchmark_config["dbconfig"], "preload_tool"
    )
    preload_tool = extract_client_tool(benchmark_config["dbconfig"], "preload_tool")
    full_benchmark_path = f"/usr/local/bin/{preload_tool}"
    client_mnt_point = "/mnt/client/"

    if "memtier_benchmark" in preload_tool:
        override_memtier_test_time_preload = 0
        (
            _,
            preload_command_str,
            arbitrary_command,
        ) = prepare_memtier_benchmark_parameters(
            benchmark_config["dbconfig"]["preload_tool"],
            full_benchmark_path,
            port,
            host,
            password,
            local_benchmark_output_filename,
            oss_cluster_api_enabled,
            tls_enabled,
            tls_skip_verify,
            tls_cert,
            tls_key,
            tls_cacert,
            resp_version,
            override_memtier_test_time_preload,
        )
        if arbitrary_command is True and oss_cluster_api_enabled:
            logging.warning(
                "Skipping this test given it implies arbitrary command on an cluster setup. Not supported on memtier: https://github.com/RedisLabs/memtier_benchmark/pull/117"
            )
            result = False
            return result

        # run the benchmark
        preload_start_time = datetime.datetime.now()

        if benchmark_local_install:
            logging.info("Running memtier benchmark outside of docker")

            preload_command_str = (
                "taskset -c " + client_cpuset_cpus + " " + preload_command_str
            )
            logging.info(
                "Pre-loading using memtier benchmark command {}".format(
                    preload_command_str
                )
            )
            stream = os.popen(preload_command_str)
            client_container_stdout = stream.read()

            move_command = "mv {} {}".format(
                local_benchmark_output_filename, temporary_dir
            )
            os.system(move_command)

        else:
            logging.info(
                "Using docker image {} as benchmark PRELOAD image (cpuset={}) with the following args: {}".format(
                    preload_image,
                    client_cpuset_cpus,
                    preload_command_str,
                )
            )
            client_container_stdout = docker_client.containers.run(
                image=preload_image,
                volumes={
                    temporary_dir: {
                        "bind": client_mnt_point,
                        "mode": "rw",
                    },
                },
                auto_remove=True,
                privileged=True,
                working_dir=benchmark_tool_workdir,
                command=preload_command_str,
                network_mode="host",
                detach=False,
                cpuset_cpus=client_cpuset_cpus,
            )

        preload_end_time = datetime.datetime.now()
        preload_duration_seconds = calculate_client_tool_duration_and_check(
            preload_end_time, preload_start_time, "Preload", False
        )
        logging.info(
            "Tool {} seconds to load data. Output {}".format(
                preload_duration_seconds,
                client_container_stdout,
            )
        )
    return result


def generate_cpuset_cpus(ceil_db_cpu_limit, current_cpu_pos):
    previous_cpu_pos = current_cpu_pos
    current_cpu_pos = current_cpu_pos + int(ceil_db_cpu_limit)
    db_cpuset_cpus = ",".join(
        [str(x) for x in range(previous_cpu_pos, current_cpu_pos)]
    )
    return db_cpuset_cpus, current_cpu_pos



import logging
from github import Github


def check_regression_comment(comments):
    res = False
    pos = -1
    for n, comment in enumerate(comments):
        body = comment.body
        if "Comparison between" in body and "Time Period from" in body:
            res = True
            pos = n
    return res, pos


def generate_build_started_pr_comment(
    build_datetime,
    commit_datetime,
    commit_summary,
    git_branch,
    git_hash,
    tests_groups_regexp,
    tests_priority_lower_limit,
    tests_priority_upper_limit,
    tests_regexp,
):
    comment_body = "### CE Performance Automation : step 1 of 2 (build) STARTING...\n\n"
    comment_body += (
        "This comment was automatically generated given a benchmark was triggered.\n"
    )
    comment_body += f"Started building at {build_datetime}\n"
    comment_body += "You can check each build/benchmark progress in grafana:\n"
    if not isinstance(git_hash, str):
        git_hash = git_hash.decode()
    comment_body += f"   - git hash: {git_hash}\n"
    comment_body += f"   - git branch: {git_branch}\n"
    comment_body += f"   - commit date and time: {commit_datetime}\n"
    comment_body += f"   - commit summary: {commit_summary}\n"
    comment_body += "   - test filters:\n"
    comment_body += (
        f"       - command priority lower limit: {tests_priority_lower_limit}\n"
    )
    comment_body += (
        f"       - command priority upper limit: {tests_priority_upper_limit}\n"
    )
    comment_body += f"       - test name regex: {tests_regexp}\n"
    comment_body += f"       - command group regex: {tests_groups_regexp}\n\n"
    return comment_body


def generate_build_finished_pr_comment(
    benchmark_stream_ids,
    commit_datetime,
    commit_summary,
    git_branch,
    git_hash,
    tests_groups_regexp,
    tests_priority_lower_limit,
    tests_priority_upper_limit,
    tests_regexp,
    build_start_datetime,
    build_duration_seconds,
):
    build_duration_seconds = int(build_duration_seconds)
    comment_body = "### CE Performance Automation : step 1 of 2 (build) DONE.\n\n"
    comment_body += (
        "This comment was automatically generated given a benchmark was triggered.\n"
    )
    comment_body += f"Started building at {build_start_datetime} and took {build_duration_seconds} seconds.\n"
    comment_body += "You can check each build/benchmark progress in grafana:\n"
    if not isinstance(git_hash, str):
        git_hash = git_hash.decode()
    comment_body += f"   - git hash: {git_hash}\n"
    comment_body += f"   - git branch: {git_branch}\n"
    comment_body += f"   - commit date and time: {commit_datetime}\n"
    comment_body += f"   - commit summary: {commit_summary}\n"
    comment_body += "   - test filters:\n"
    comment_body += (
        f"       - command priority lower limit: {tests_priority_lower_limit}\n"
    )
    comment_body += (
        f"       - command priority upper limit: {tests_priority_upper_limit}\n"
    )
    comment_body += f"       - test name regex: {tests_regexp}\n"
    comment_body += f"       - command group regex: {tests_groups_regexp}\n\n"
    for benchmark_stream_id in benchmark_stream_ids:
        if not isinstance(benchmark_stream_id, str):
            benchmark_stream_id = benchmark_stream_id.decode()
        grafana_benchmark_status_link = f"https://benchmarksredisio.grafana.net/d/edsxdsrbexhc0f/ce-benchmark-run-status?orgId=1&var-benchmark_work_stream={benchmark_stream_id}"
        print("=============================================================")
        print(f"Check benchmark run status in: {grafana_benchmark_status_link}")
        comment_body += f"You can check a comparison in detail via the [grafana link]({grafana_benchmark_status_link})"
    return comment_body


def check_github_available_and_actionable(
    fn, github_token, pull_request, tf_github_org, tf_github_repo, verbose
):
    # using an access token
    is_actionable_pr = False
    contains_regression_comment = False
    regression_comment = None
    github_pr = None
    old_regression_comment_body = ""
    pr_link = None
    if github_token is not None:
        logging.info("Detected github token")
        g = Github(github_token)
        if pull_request is not None and pull_request != "":
            pull_request_n = int(pull_request)
            github_pr = (
                g.get_user(tf_github_org)
                .get_repo(tf_github_repo)
                .get_issue(pull_request_n)
            )
            comments = github_pr.get_comments()
            pr_link = github_pr.html_url
            logging.info("Working on github PR already: {}".format(pr_link))
            is_actionable_pr = True
            contains_regression_comment, pos = fn(comments)
            if contains_regression_comment:
                regression_comment = comments[pos]
                old_regression_comment_body = regression_comment.body
                logging.info(
                    "Already contains PR comment. Link: {}".format(
                        regression_comment.html_url
                    )
                )
                if verbose:
                    logging.info("Printing old PR comment:")
                    print("".join(["-" for x in range(1, 80)]))
                    print(regression_comment.body)
                    print("".join(["-" for x in range(1, 80)]))
            else:
                logging.info("Does not contain PR comment")
            logging.info(
                f"contains_regression_comment: {contains_regression_comment}, is_actionable_pr: {is_actionable_pr}, pr_link: {pr_link}"
            )

    return (
        contains_regression_comment,
        github_pr,
        is_actionable_pr,
        old_regression_comment_body,
        pr_link,
        regression_comment,
    )


def create_new_pr_comment(auto_approve, comment_body, github_pr, pr_link):
    regression_comment = None
    user_input = "n"
    if auto_approve:
        print("auto approving...")
    else:
        user_input = input("Do you wish to add a comment in {} (y/n): ".format(pr_link))
    if user_input.lower() == "y" or auto_approve:
        print("creating an comment in PR {}".format(pr_link))
        regression_comment = github_pr.create_comment(comment_body)
        html_url = regression_comment.html_url
        print("created comment. Access it via {}".format(html_url))
    return regression_comment


def update_comment_if_needed(
    auto_approve, comment_body, old_regression_comment_body, regression_comment, verbose
):
    same_comment = False
    user_input = "n"
    if comment_body == old_regression_comment_body:
        logging.info(
            "The old github comment is the same as the new comment. skipping..."
        )
        same_comment = True
    else:
        logging.info(
            "The old github comment is different from the new comment. updating it..."
        )
        comment_body_arr = comment_body.split("\n")
        old_regression_comment_body_arr = old_regression_comment_body.split("\n")
        if verbose:
            DF = [
                x for x in comment_body_arr if x not in old_regression_comment_body_arr
            ]
            print("---------------------")
            print(DF)
            print("---------------------")
    if same_comment is False:
        if auto_approve:
            print("auto approving...")
        else:
            user_input = input(
                "Do you wish to update the comment {} (y/n): ".format(
                    regression_comment.html_url
                )
            )
        if user_input.lower() == "y" or auto_approve:
            html_url = regression_comment.html_url
            print("Updating comment {}".format(html_url))
            regression_comment.edit(comment_body)
            print("Updated comment. Access it via {}".format(html_url))


def check_benchmark_build_comment(comments):
    res = False
    pos = -1
    for n, comment in enumerate(comments):
        body = comment.body
        if "CE Performance Automation : step" in body:
            res = True
            pos = n
    return res, pos


def check_benchmark_running_comment(comments):
    res = False
    pos = -1
    for n, comment in enumerate(comments):
        body = comment.body
        if "CE Performance Automation : step" in body:
            res = True
            pos = n
    return res, pos


def markdown_progress_bar(current, total, bar_length=40):
    progress = 1.0
    if total > 0:
        progress = current / total
    block = int(round(bar_length * progress))
    bar = "#" * block + "-" * (bar_length - block)
    percentage = round(progress * 100, 2)
    return f"[{bar}] {percentage}%"


def generate_benchmark_started_pr_comment(
    benchmark_stream_id,
    total_pending,
    total_benchmarks,
    total_failed,
    benchmark_suite_start_datetime,
    benchmark_suite_duration_secs,
):
    comment_body = "### CE Performance Automation : step 2 of 2 (benchmark) "
    if total_pending > 0:
        comment_body += "RUNNING...\n\n"
    else:
        comment_body += "FINISHED.\n\n"

    comment_body += (
        "This comment was automatically generated given a benchmark was triggered.\n\n"
    )

    comment_body += f"Started benchmark suite at {benchmark_suite_start_datetime} and took {benchmark_suite_duration_secs} seconds "
    if total_pending > 0:
        comment_body += "up until now.\n"
    else:
        comment_body += "to finish.\n"

    completed = total_benchmarks - total_pending
    successful = completed - total_failed
    comment_body += (
        f"Status: {markdown_progress_bar(completed,total_benchmarks,80)} completed.\n\n"
    )
    comment_body += f"In total will run {total_benchmarks} benchmarks.\n"
    comment_body += f"    - {total_pending} pending.\n"
    comment_body += f"    - {completed} completed:\n"
    comment_body += f"      - {successful} successful.\n"
    comment_body += f"      - {total_failed} failed.\n"

    if not isinstance(benchmark_stream_id, str):
        benchmark_stream_id = benchmark_stream_id.decode()
    grafana_benchmark_status_link = f"https://benchmarksredisio.grafana.net/d/edsxdsrbexhc0f/ce-benchmark-run-status?orgId=1&var-benchmark_work_stream={benchmark_stream_id}"
    comment_body += f"You can check a the status in detail via the [grafana link]({grafana_benchmark_status_link})"
    return comment_body



import math
import logging


def extract_redis_dbconfig_parameters(benchmark_config, dbconfig_keyname):
    redis_configuration_parameters = {}
    modules_configuration_parameters_map = {}
    dataset_load_timeout_secs = 120
    dataset_name = None
    dbconfig_present = False
    if dbconfig_keyname in benchmark_config:
        dbconfig_present = True
        if type(benchmark_config[dbconfig_keyname]) == list:
            for k in benchmark_config[dbconfig_keyname]:
                if "configuration-parameters" in k:
                    cp = k["configuration-parameters"]
                    for item in cp:
                        for k, v in item.items():
                            redis_configuration_parameters[k] = v
                if "dataset_load_timeout_secs" in k:
                    dataset_load_timeout_secs = k["dataset_load_timeout_secs"]
                if "dataset_name" in k:
                    dataset_name = k["dataset_name"]
        if type(benchmark_config[dbconfig_keyname]) == dict:
            if "configuration-parameters" in benchmark_config[dbconfig_keyname]:
                cp = benchmark_config[dbconfig_keyname]["configuration-parameters"]
                for k, v in cp.items():
                    redis_configuration_parameters[k] = v
            if "dataset_load_timeout_secs" in benchmark_config[dbconfig_keyname]:
                dataset_load_timeout_secs = benchmark_config[dbconfig_keyname][
                    "dataset_load_timeout_secs"
                ]
            if "dataset_name" in benchmark_config[dbconfig_keyname]:
                dataset_name = benchmark_config[dbconfig_keyname]["dataset_name"]

    return (
        dbconfig_present,
        dataset_name,
        redis_configuration_parameters,
        dataset_load_timeout_secs,
        modules_configuration_parameters_map,
    )


def extract_redis_configuration_from_topology(topologies_map, topology_spec_name):
    redis_arguments = ""
    topology_spec = topologies_map[topology_spec_name]
    if "redis_arguments" in topology_spec:
        redis_arguments = topology_spec["redis_arguments"]
        logging.info(
            f"extracted redis_arguments: {redis_arguments} from topology: {topology_spec_name}"
        )
    return redis_arguments


def extract_client_cpu_limit(benchmark_config):
    db_cpu_limit = benchmark_config["clientconfig"]["resources"]["requests"]["cpus"]
    ceil_db_cpu_limit = math.ceil(float(db_cpu_limit))
    return ceil_db_cpu_limit


def extract_build_variant_variations(benchmark_config, keyname="build-variants"):
    result = None
    if keyname in benchmark_config:
        result = benchmark_config[keyname]
    return result


def extract_client_container_image(benchmark_config, keyname="clientconfig"):
    client_container_image = None
    if keyname in benchmark_config:
        if "run_image" in benchmark_config[keyname]:
            client_container_image = benchmark_config[keyname]["run_image"]
    return client_container_image


def extract_client_tool(benchmark_config, keyname="clientconfig"):
    client_tool = None
    if keyname in benchmark_config:
        if "tool" in benchmark_config[keyname]:
            client_tool = benchmark_config[keyname]["tool"]
    return client_tool



#  BSD 3-Clause License
#
#  Copyright (c) 2021., Redis Labs Modules
#  All rights reserved.
#
import datetime
import logging
from tqdm import tqdm
import redis
from jsonpath_ng.parser import JsonPathParser


def parse(string):
    return JsonPathParser().parse(string)


def parse_exporter_timemetric(metric_path: str, results_dict: dict):
    datapoints_timestamp = None
    try:
        jsonpath_expr = parse(metric_path)
        find_res = jsonpath_expr.find(results_dict)
        if len(find_res) > 0:
            datapoints_timestamp = int(find_res[0].value)
    except Exception as e:
        logging.error(
            "Unable to parse time-metric {}. Error: {}".format(metric_path, e.__str__())
        )
    return datapoints_timestamp


def parse_exporter_timemetric_definition(
    benchmark_config: dict, configkey: str = "redistimeseries"
):
    metric_path = None
    if "timemetric" in benchmark_config[configkey]:
        metric_path = benchmark_config[configkey]["timemetric"]
    return metric_path


def parse_exporter_metrics_definition(
    benchmark_config: dict, configkey: str = "redistimeseries"
):
    metrics = []
    if configkey in benchmark_config:
        if "metrics" in benchmark_config[configkey]:
            for metric_name in benchmark_config[configkey]["metrics"]:
                metrics.append(metric_name)
    return metrics


def get_ts_metric_name(
    by,
    by_value,
    tf_github_org,
    tf_github_repo,
    deployment_name,
    deployment_type,
    test_name,
    tf_triggering_env,
    metric_name,
    metric_context_path=None,
    use_metric_context_path=False,
    build_variant_name=None,
    running_platform=None,
):
    if use_metric_context_path:
        metric_name = "{}/{}".format(metric_name, metric_context_path)
    build_variant_str = ""
    if build_variant_name is not None:
        build_variant_str = "{}/".format(str(build_variant_name))
    running_platform_str = ""
    if running_platform is not None:
        running_platform_str = "{}/".format(str(running_platform))
    if deployment_name != deployment_type:
        deployment_name = "/{}".format(deployment_name)
    else:
        deployment_name = ""
    ts_name = (
        "ci.benchmarks.redislabs/{by}/"
        "{triggering_env}/{github_org}/{github_repo}/"
        "{test_name}/{build_variant_str}{running_platform_str}{deployment_type}{deployment_name}/{by_value}/{metric}".format(
            by=by,
            triggering_env=tf_triggering_env,
            github_org=tf_github_org,
            github_repo=tf_github_repo,
            test_name=test_name,
            deployment_type=deployment_type,
            deployment_name=deployment_name,
            build_variant_str=build_variant_str,
            running_platform_str=running_platform_str,
            by_value=str(by_value),
            metric=metric_name,
        )
    )
    return ts_name


def extract_results_table(
    metrics,
    results_dict,
):
    results_matrix = []
    cleaned_metrics = []
    already_present_metrics = []
    # insert first the dict metrics
    for jsonpath in metrics:
        if type(jsonpath) == dict:
            cleaned_metrics.append(jsonpath)
            metric_jsonpath = list(jsonpath.keys())[0]
            already_present_metrics.append(metric_jsonpath)
    for jsonpath in metrics:
        if type(jsonpath) == str:
            if jsonpath not in already_present_metrics:
                already_present_metrics.append(jsonpath)
                cleaned_metrics.append(jsonpath)

    for jsonpath in cleaned_metrics:
        test_case_targets_dict = {}
        metric_jsonpath = jsonpath
        find_res = None
        try:
            if type(jsonpath) == str:
                jsonpath_expr = parse(jsonpath)
            if type(jsonpath) == dict:
                metric_jsonpath = list(jsonpath.keys())[0]
                test_case_targets_dict = jsonpath[metric_jsonpath]
                jsonpath_expr = parse(metric_jsonpath)
            find_res = jsonpath_expr.find(results_dict)
        except Exception:
            pass
        finally:
            if find_res is not None:
                use_metric_context_path = False
                if len(find_res) > 1:
                    use_metric_context_path = True
                for metric in find_res:
                    metric_name = str(metric.path)
                    metric_value = float(metric.value)
                    metric_context_path = str(metric.context.path)
                    if metric_jsonpath[0] == "$":
                        metric_jsonpath = metric_jsonpath[1:]
                    if metric_jsonpath[0] == ".":
                        metric_jsonpath = metric_jsonpath[1:]

                    # retro-compatible naming
                    if use_metric_context_path is False:
                        metric_name = metric_jsonpath

                    metric_name = metric_name.replace("'", "")
                    metric_name = metric_name.replace('"', "")
                    metric_name = metric_name.replace("(", "")
                    metric_name = metric_name.replace(")", "")
                    metric_name = metric_name.replace(" ", "_")

                    results_matrix.append(
                        [
                            metric_jsonpath,
                            metric_context_path,
                            metric_name,
                            metric_value,
                            test_case_targets_dict,
                            use_metric_context_path,
                        ]
                    )

            else:
                logging.warning(
                    "Unable to find metric path {} in result dict".format(jsonpath)
                )
    return results_matrix


def get_ts_tags_and_name(
    break_by_key,
    break_by_str,
    break_by_value,
    build_variant_name,
    deployment_name,
    deployment_type,
    metadata_tags,
    metric_context_path,
    metric_jsonpath,
    metric_name,
    running_platform,
    test_name,
    testcase_metric_context_paths,
    tf_github_org,
    tf_github_repo,
    tf_triggering_env,
    use_metric_context_path,
):
    # prepare tags
    timeserie_tags = get_project_ts_tags(
        tf_github_org,
        tf_github_repo,
        deployment_name,
        deployment_type,
        tf_triggering_env,
        metadata_tags,
        build_variant_name,
        running_platform,
    )
    timeserie_tags[break_by_key] = break_by_value
    timeserie_tags["{}+{}".format("deployment_name", break_by_key)] = "{} {}".format(
        deployment_name, break_by_value
    )
    timeserie_tags[break_by_key] = break_by_value
    timeserie_tags["{}+{}".format("target", break_by_key)] = "{} {}".format(
        break_by_value, tf_github_repo
    )
    timeserie_tags["test_name"] = str(test_name)
    if build_variant_name is not None:
        timeserie_tags["test_name:build_variant"] = "{}:{}".format(
            test_name, build_variant_name
        )
    timeserie_tags["metric"] = str(metric_name)
    timeserie_tags["metric_name"] = metric_name
    timeserie_tags["metric_context_path"] = metric_context_path
    if metric_context_path is not None:
        timeserie_tags["test_name:metric_context_path"] = "{}:{}".format(
            test_name, metric_context_path
        )
    timeserie_tags["metric_jsonpath"] = metric_jsonpath
    if metric_context_path not in testcase_metric_context_paths:
        testcase_metric_context_paths.append(metric_context_path)
    ts_name = get_ts_metric_name(
        break_by_str,
        break_by_value,
        tf_github_org,
        tf_github_repo,
        deployment_name,
        deployment_type,
        test_name,
        tf_triggering_env,
        metric_name,
        metric_context_path,
        use_metric_context_path,
        build_variant_name,
        running_platform,
    )
    return timeserie_tags, ts_name


def from_metric_kv_to_timeserie(
    break_by_key,
    break_by_str,
    break_by_value,
    build_variant_name,
    datapoints_timestamp,
    deployment_name,
    deployment_type,
    metadata_tags,
    metric_context_path,
    metric_jsonpath,
    metric_name,
    metric_value,
    running_platform,
    test_case_targets_dict,
    test_name,
    testcase_metric_context_paths,
    tf_github_org,
    tf_github_repo,
    tf_triggering_env,
    time_series_dict,
    use_metric_context_path,
):
    timeserie_tags, ts_name = get_ts_tags_and_name(
        break_by_key,
        break_by_str,
        break_by_value,
        build_variant_name,
        deployment_name,
        deployment_type,
        metadata_tags,
        metric_context_path,
        metric_jsonpath,
        metric_name,
        running_platform,
        test_name,
        testcase_metric_context_paths,
        tf_github_org,
        tf_github_repo,
        tf_triggering_env,
        use_metric_context_path,
    )
    logging.info(f"Adding timeserie named {ts_name} to time_series_dict.")
    time_series_dict[ts_name] = {
        "labels": timeserie_tags.copy(),
        "data": {datapoints_timestamp: metric_value},
    }

    original_ts_name = ts_name
    target_table_keyname = "target_tables:{triggering_env}:ci.benchmarks.redislabs/{break_by_key}/{break_by_str}/{tf_github_org}/{tf_github_repo}/{deployment_type}/{deployment_name}/{test_name}/{metric_name}".format(
        triggering_env=tf_triggering_env,
        break_by_key=break_by_key,
        break_by_str=break_by_str,
        tf_github_org=tf_github_org,
        tf_github_repo=tf_github_repo,
        deployment_name=deployment_name,
        deployment_type=deployment_type,
        test_name=test_name,
        metric_name=metric_name,
    )
    target_table_dict = {
        "test-case": test_name,
        "metric-name": metric_name,
        tf_github_repo: metric_value,
        "contains-target": False,
    }
    for target_name, target_value in test_case_targets_dict.items():
        target_table_dict["contains-target"] = True
        ts_name = original_ts_name + "/target/{}".format(target_name)
        timeserie_tags_target = timeserie_tags.copy()
        timeserie_tags_target["is_target"] = "true"
        timeserie_tags_target["{}+{}".format("target", break_by_key)] = "{} {}".format(
            break_by_value, target_name
        )
        time_series_dict[ts_name] = {
            "labels": timeserie_tags_target,
            "data": {datapoints_timestamp: target_value},
        }
        if "overallQuantiles" in metric_name:
            comparison_type = "(lower-better)"
        else:
            comparison_type = "(higher-better)"
        if comparison_type == "(higher-better)":
            target_value_pct = (
                (float(metric_value) / float(target_value)) - 1.0
            ) * 100.0
        else:
            target_value_pct = (
                (float(target_value) / float(metric_value)) - 1.0
            ) * 100.0

        target_value_pct_str = "{:.2f}".format(target_value_pct)

        target_table_dict[target_name] = target_value

        target_table_dict["{}:percent {}".format(target_name, comparison_type)] = (
            target_value_pct_str
        )
    return target_table_keyname, target_table_dict


def common_timeseries_extraction(
    break_by_key,
    break_by_str,
    datapoints_timestamp,
    deployment_name,
    deployment_type,
    metrics,
    break_by_value,
    results_dict,
    test_name,
    tf_github_org,
    tf_github_repo,
    tf_triggering_env,
    metadata_tags={},
    build_variant_name=None,
    running_platform=None,
    testcase_metric_context_paths=[],
):
    time_series_dict = {}
    target_tables = {}
    cleaned_metrics_arr = extract_results_table(metrics, results_dict)
    total_metrics = len(cleaned_metrics_arr)
    logging.info(f"Total of {total_metrics} cleaned metrics: {cleaned_metrics_arr}")
    for cleaned_metric in cleaned_metrics_arr:

        metric_jsonpath = cleaned_metric[0]
        metric_context_path = cleaned_metric[1]
        metric_name = cleaned_metric[2]
        metric_value = cleaned_metric[3]
        test_case_targets_dict = cleaned_metric[4]
        use_metric_context_path = cleaned_metric[5]

        target_table_keyname, target_table_dict = from_metric_kv_to_timeserie(
            break_by_key,
            break_by_str,
            break_by_value,
            build_variant_name,
            datapoints_timestamp,
            deployment_name,
            deployment_type,
            metadata_tags,
            metric_context_path,
            metric_jsonpath,
            metric_name,
            metric_value,
            running_platform,
            test_case_targets_dict,
            test_name,
            testcase_metric_context_paths,
            tf_github_org,
            tf_github_repo,
            tf_triggering_env,
            time_series_dict,
            use_metric_context_path,
        )
        target_tables[target_table_keyname] = target_table_dict

    return time_series_dict, target_tables


def extract_perhash_timeseries_from_results(
    datapoints_timestamp: int,
    metrics: list,
    results_dict: dict,
    git_hash: str,
    tf_github_org: str,
    tf_github_repo: str,
    deployment_name: str,
    deployment_type: str,
    test_name: str,
    tf_triggering_env: str,
    metadata_tags={},
    build_variant_name=None,
    running_platform=None,
    testcase_metric_context_paths=[],
):
    break_by_key = "hash"
    break_by_str = "by.{}".format(break_by_key)
    (
        time_series_dict,
        target_tables,
    ) = common_timeseries_extraction(
        break_by_key,
        break_by_str,
        datapoints_timestamp,
        deployment_name,
        deployment_type,
        metrics,
        git_hash,
        results_dict,
        test_name,
        tf_github_org,
        tf_github_repo,
        tf_triggering_env,
        metadata_tags,
        build_variant_name,
        running_platform,
        testcase_metric_context_paths,
    )
    return True, time_series_dict, target_tables


def extract_perversion_timeseries_from_results(
    datapoints_timestamp: int,
    metrics: list,
    results_dict: dict,
    project_version: str,
    tf_github_org: str,
    tf_github_repo: str,
    deployment_name: str,
    deployment_type: str,
    test_name: str,
    tf_triggering_env: str,
    metadata_tags={},
    build_variant_name=None,
    running_platform=None,
    testcase_metric_context_paths=[],
):
    break_by_key = "version"
    break_by_str = "by.{}".format(break_by_key)
    (
        branch_time_series_dict,
        target_tables,
    ) = common_timeseries_extraction(
        break_by_key,
        break_by_str,
        datapoints_timestamp,
        deployment_name,
        deployment_type,
        metrics,
        project_version,
        results_dict,
        test_name,
        tf_github_org,
        tf_github_repo,
        tf_triggering_env,
        metadata_tags,
        build_variant_name,
        running_platform,
        testcase_metric_context_paths,
    )
    return True, branch_time_series_dict, target_tables


def push_data_to_redistimeseries(rts, time_series_dict: dict, expire_msecs=0):
    datapoint_errors = 0
    datapoint_inserts = 0
    if rts is not None and time_series_dict is not None:
        progress = tqdm(
            unit="benchmark time-series", total=len(time_series_dict.values())
        )
        for timeseries_name, time_series in time_series_dict.items():
            exporter_create_ts(rts, time_series, timeseries_name)
            for timestamp, value in time_series["data"].items():
                try:
                    if timestamp is None:
                        logging.warning("The provided timestamp is null. Using auto-ts")
                        rts.ts().add(
                            timeseries_name,
                            value,
                            duplicate_policy="last",
                        )
                    else:
                        rts.ts().add(
                            timeseries_name,
                            timestamp,
                            value,
                            duplicate_policy="last",
                        )
                    datapoint_inserts += 1
                except redis.exceptions.DataError:
                    logging.warning(
                        "Error while inserting datapoint ({} : {}) in timeseries named {}. ".format(
                            timestamp, value, timeseries_name
                        )
                    )
                    datapoint_errors += 1
                    pass
                except redis.exceptions.ResponseError:
                    logging.warning(
                        "Error while inserting datapoint ({} : {}) in timeseries named {}. ".format(
                            timestamp, value, timeseries_name
                        )
                    )
                    datapoint_errors += 1
                    pass
            if expire_msecs > 0:
                rts.pexpire(timeseries_name, expire_msecs)
            progress.update()
    return datapoint_errors, datapoint_inserts


def extract_perbranch_timeseries_from_results(
    datapoints_timestamp: int,
    metrics: list,
    results_dict: dict,
    tf_github_branch: str,
    tf_github_org: str,
    tf_github_repo: str,
    deployment_name: str,
    deployment_type: str,
    test_name: str,
    tf_triggering_env: str,
    metadata_tags={},
    build_variant_name=None,
    running_platform=None,
    testcase_metric_context_paths=[],
):
    break_by_key = "branch"
    break_by_str = "by.{}".format(break_by_key)
    (branch_time_series_dict, target_tables) = common_timeseries_extraction(
        break_by_key,
        break_by_str,
        datapoints_timestamp,
        deployment_name,
        deployment_type,
        metrics,
        tf_github_branch,
        results_dict,
        test_name,
        tf_github_org,
        tf_github_repo,
        tf_triggering_env,
        metadata_tags,
        build_variant_name,
        running_platform,
        testcase_metric_context_paths,
    )
    return True, branch_time_series_dict, target_tables


def check_rts_labels(rts, time_series, timeseries_name):
    updated_create = False
    logging.debug(
        "Timeseries named {} already exists. Checking that the labels match.".format(
            timeseries_name
        )
    )
    set1 = set(time_series["labels"].items())
    set2 = set(rts.ts().info(timeseries_name).labels.items())
    if len(set1 - set2) > 0 or len(set2 - set1) > 0:
        logging.info(
            "Given the labels don't match using TS.ALTER on {} to update labels to {}".format(
                timeseries_name, time_series["labels"]
            )
        )
        updated_create = True
        rts.ts().alter(timeseries_name, labels=time_series["labels"])
    return updated_create


def exporter_create_ts(rts, time_series, timeseries_name):
    updated_create = False
    try:
        if rts.exists(timeseries_name):
            updated_create = check_rts_labels(rts, time_series, timeseries_name)
        else:
            logging.debug(
                "Creating timeseries named {} with labels {}".format(
                    timeseries_name, time_series["labels"]
                )
            )
            rts.ts().create(
                timeseries_name, labels=time_series["labels"], chunk_size=128
            )
            updated_create = True

    except redis.exceptions.ResponseError as e:
        if "already exists" in e.__str__():
            updated_create = check_rts_labels(rts, time_series, timeseries_name)
            pass
        else:
            logging.error(
                "While creating timeseries named {} with the following labels: {} this error ocurred: {}".format(
                    timeseries_name, time_series["labels"], e.__str__()
                )
            )
            raise
    return updated_create


def get_overall_dashboard_keynames(
    tf_github_org,
    tf_github_repo,
    tf_triggering_env,
    build_variant_name=None,
    running_platform=None,
    test_name=None,
):
    build_variant_str = ""
    if build_variant_name is not None:
        build_variant_str = "/{}".format(build_variant_name)
    running_platform_str = ""
    if running_platform is not None:
        running_platform_str = "/{}".format(running_platform)
    sprefix = (
        "ci.benchmarks.redislabs/"
        + "{triggering_env}/{github_org}/{github_repo}".format(
            triggering_env=tf_triggering_env,
            github_org=tf_github_org,
            github_repo=tf_github_repo,
        )
    )
    testcases_setname = "{}:testcases".format(sprefix)
    deployment_name_setname = "{}:deployment_names".format(sprefix)
    project_archs_setname = "{}:archs".format(sprefix)
    project_oss_setname = "{}:oss".format(sprefix)
    project_branches_setname = "{}:branches".format(sprefix)
    project_versions_setname = "{}:versions".format(sprefix)
    project_compilers_setname = "{}:compilers".format(sprefix)
    running_platforms_setname = "{}:platforms".format(sprefix)
    build_variant_setname = "{}:build_variants".format(sprefix)
    build_variant_prefix = "{sprefix}{build_variant_str}".format(
        sprefix=sprefix,
        build_variant_str=build_variant_str,
    )
    prefix = "{build_variant_prefix}{running_platform_str}".format(
        build_variant_prefix=build_variant_prefix,
        running_platform_str=running_platform_str,
    )
    tsname_project_total_success = "{}:total_success".format(
        prefix,
    )
    tsname_project_total_failures = "{}:total_failures".format(
        prefix,
    )
    testcases_metric_context_path_setname = ""
    if test_name is not None:
        testcases_metric_context_path_setname = (
            "{testcases_setname}:metric_context_path:{test_name}".format(
                testcases_setname=testcases_setname, test_name=test_name
            )
        )
    testcases_and_metric_context_path_setname = (
        "{testcases_setname}_AND_metric_context_path".format(
            testcases_setname=testcases_setname
        )
    )
    return (
        prefix,
        testcases_setname,
        deployment_name_setname,
        tsname_project_total_failures,
        tsname_project_total_success,
        running_platforms_setname,
        build_variant_setname,
        testcases_metric_context_path_setname,
        testcases_and_metric_context_path_setname,
        project_archs_setname,
        project_oss_setname,
        project_branches_setname,
        project_versions_setname,
        project_compilers_setname,
    )


def get_project_ts_tags(
    tf_github_org: str,
    tf_github_repo: str,
    deployment_name: str,
    deployment_type: str,
    tf_triggering_env: str,
    metadata_tags={},
    build_variant_name=None,
    running_platform=None,
):
    tags = {
        "github_org": tf_github_org,
        "github_repo": tf_github_repo,
        "github_org/github_repo": "{}/{}".format(tf_github_org, tf_github_repo),
        "deployment_type": deployment_type,
        "deployment_name": deployment_name,
        "triggering_env": tf_triggering_env,
    }
    if build_variant_name is not None:
        tags["build_variant"] = build_variant_name
    if running_platform is not None:
        tags["running_platform"] = running_platform
    for k, v in metadata_tags.items():
        tags[k] = str(v)
    return tags


def common_exporter_logic(
    deployment_name,
    deployment_type,
    exporter_timemetric_path,
    metrics,
    results_dict,
    test_name,
    tf_github_branch,
    tf_github_org,
    tf_github_repo,
    tf_triggering_env,
    artifact_version="N/A",
    metadata_tags={},
    build_variant_name=None,
    running_platform=None,
    datapoints_timestamp=None,
    git_hash=None,
):
    per_version_time_series_dict = {}
    per_branch_time_series_dict = {}
    per_hash_time_series_dict = {}
    testcase_metric_context_paths = []
    version_target_tables = None
    branch_target_tables = None
    hash_target_tables = None
    used_ts = datapoints_timestamp

    if exporter_timemetric_path is not None and used_ts is None:
        # extract timestamp
        used_ts = parse_exporter_timemetric(exporter_timemetric_path, results_dict)

    if used_ts is None:
        used_ts = int(datetime.datetime.now(datetime.timezone.utc).timestamp() * 1000.0)
        logging.warning(
            "Error while trying to parse datapoints timestamp. Using current system timestamp Error: {}".format(
                used_ts
            )
        )
    assert used_ts is not None
    total_break_by_added = 0
    if (git_hash is not None) and (git_hash != ""):
        # extract per-hash datapoints
        (
            _,
            per_hash_time_series_dict,
            version_target_tables,
        ) = extract_perhash_timeseries_from_results(
            used_ts,
            metrics,
            results_dict,
            git_hash,
            tf_github_org,
            tf_github_repo,
            deployment_name,
            deployment_type,
            test_name,
            tf_triggering_env,
            metadata_tags,
            build_variant_name,
            running_platform,
            testcase_metric_context_paths,
        )
        total_break_by_added += 1
    else:
        logging.warning(
            "there was no git hash information to push data brokedown by hash"
        )
    if (
        artifact_version is not None
        and artifact_version != ""
        and artifact_version != "N/A"
    ):
        # extract per-version datapoints
        total_hs_ts = len(per_hash_time_series_dict.keys())
        logging.info(
            f"Extending the by.hash {git_hash} timeseries ({total_hs_ts}) with version info {artifact_version}"
        )
        for hash_timeserie in per_hash_time_series_dict.values():
            hash_timeserie["labels"]["version"] = artifact_version
        (
            _,
            per_version_time_series_dict,
            version_target_tables,
        ) = extract_perversion_timeseries_from_results(
            used_ts,
            metrics,
            results_dict,
            artifact_version,
            tf_github_org,
            tf_github_repo,
            deployment_name,
            deployment_type,
            test_name,
            tf_triggering_env,
            metadata_tags,
            build_variant_name,
            running_platform,
            testcase_metric_context_paths,
        )
        total_break_by_added += 1
    else:
        logging.warning(
            "there was no git VERSION information to push data brokedown by VERSION"
        )
    if tf_github_branch is not None and tf_github_branch != "":
        total_hs_ts = len(per_hash_time_series_dict.keys())
        logging.info(
            f"Extending the by.hash {git_hash} timeseries ({total_hs_ts}) with branch info {tf_github_branch}"
        )
        for hash_timeserie in per_hash_time_series_dict.values():
            hash_timeserie["labels"]["branch"] = tf_github_branch
        # extract per branch datapoints
        (
            _,
            per_branch_time_series_dict,
            branch_target_tables,
        ) = extract_perbranch_timeseries_from_results(
            used_ts,
            metrics,
            results_dict,
            str(tf_github_branch),
            tf_github_org,
            tf_github_repo,
            deployment_name,
            deployment_type,
            test_name,
            tf_triggering_env,
            metadata_tags,
            build_variant_name,
            running_platform,
            testcase_metric_context_paths,
        )
        total_break_by_added += 1
    else:
        logging.warning(
            "there was no git BRANCH information to push data brokedown by BRANCH"
        )
    if total_break_by_added == 0:
        logging.error(
            "There was no BRANCH, HASH, or VERSION info to break this info by in timeseries"
        )
    return (
        per_version_time_series_dict,
        per_branch_time_series_dict,
        per_hash_time_series_dict,
        testcase_metric_context_paths,
        version_target_tables,
        branch_target_tables,
        hash_target_tables,
    )


def merge_default_and_config_metrics(
    benchmark_config, default_metrics, exporter_timemetric_path
):
    if default_metrics is None:
        default_metrics = []
    metrics = default_metrics
    if benchmark_config is not None:
        if "exporter" in benchmark_config:
            extra_metrics = parse_exporter_metrics_definition(
                benchmark_config["exporter"]
            )
            metrics.extend(extra_metrics)
            extra_timemetric_path = parse_exporter_timemetric_definition(
                benchmark_config["exporter"]
            )
            if extra_timemetric_path is not None:
                exporter_timemetric_path = extra_timemetric_path
    return exporter_timemetric_path, metrics


def get_profilers_rts_key_prefix(triggering_env, tf_github_org, tf_github_repo):
    zset_name = "ci.benchmarks.redis.com/{triggering_env}/{github_org}/{github_repo}:profiles".format(
        triggering_env=triggering_env,
        github_org=tf_github_org,
        github_repo=tf_github_repo,
    )
    return zset_name


def prepare_timeseries_dict(
    artifact_version,
    benchmark_config,
    default_metrics,
    deployment_name,
    deployment_type,
    exporter_timemetric_path,
    results_dict,
    test_name,
    tf_github_branch,
    tf_github_org,
    tf_github_repo,
    tf_triggering_env,
    metadata_tags={},
    build_variant_name=None,
    running_platform=None,
    datapoints_timestamp=None,
    git_hash=None,
):
    time_series_dict = {}
    # check which metrics to extract
    exporter_timemetric_path, metrics = merge_default_and_config_metrics(
        benchmark_config, default_metrics, exporter_timemetric_path
    )
    (
        per_version_time_series_dict,
        per_branch_time_series_dict,
        per_hash_timeseries_dict,
        testcase_metric_context_paths,
        version_target_tables,
        branch_target_tables,
        _,
    ) = common_exporter_logic(
        deployment_name,
        deployment_type,
        exporter_timemetric_path,
        metrics,
        results_dict,
        test_name,
        tf_github_branch,
        tf_github_org,
        tf_github_repo,
        tf_triggering_env,
        artifact_version,
        metadata_tags,
        build_variant_name,
        running_platform,
        datapoints_timestamp,
        git_hash,
    )
    time_series_dict.update(per_version_time_series_dict)
    time_series_dict.update(per_branch_time_series_dict)
    time_series_dict.update(per_hash_timeseries_dict)
    return (
        time_series_dict,
        testcase_metric_context_paths,
        version_target_tables,
        branch_target_tables,
    )


def add_standardized_metric_bybranch(
    metric_name,
    metric_value,
    tf_github_branch,
    deployment_name,
    deployment_type,
    rts,
    start_time_ms,
    test_name,
    tf_github_org,
    tf_github_repo,
    tf_triggering_env,
    metadata_tags={},
    build_variant_name=None,
    running_platform=None,
):
    if metric_value is not None:
        tsname_use_case_duration = get_ts_metric_name(
            "by.branch",
            tf_github_branch,
            tf_github_org,
            tf_github_repo,
            deployment_name,
            deployment_type,
            test_name,
            tf_triggering_env,
            metric_name,
            None,
            False,
            build_variant_name,
            running_platform,
        )
        labels = get_project_ts_tags(
            tf_github_org,
            tf_github_repo,
            deployment_name,
            deployment_type,
            tf_triggering_env,
            metadata_tags,
            build_variant_name,
            running_platform,
        )
        labels["branch"] = tf_github_branch
        labels["deployment_name+branch"] = "{} {}".format(
            deployment_name, tf_github_branch
        )
        labels["test_name"] = str(test_name)
        labels["metric"] = str(metric_name)
        logging.info(
            "Adding metric {}={} to time-serie named {}".format(
                metric_name, metric_value, tsname_use_case_duration
            )
        )
        ts = {"labels": labels}
        exporter_create_ts(rts, ts, tsname_use_case_duration)
        logging.error(labels)
        rts.ts().add(
            tsname_use_case_duration,
            start_time_ms,
            metric_value,
            labels=labels,
        )
    else:
        logging.warning(
            "Given that metric {}={} ( is None ) we will skip adding it to timeseries".format(
                metric_name, metric_value
            )
        )


def add_standardized_metric_byversion(
    metric_name,
    metric_value,
    artifact_version,
    deployment_name,
    deployment_type,
    rts,
    start_time_ms,
    test_name,
    tf_github_org,
    tf_github_repo,
    tf_triggering_env,
    metadata_tags={},
    build_variant_name=None,
    running_platform=None,
):
    if metric_value is not None:
        tsname_use_case_duration = get_ts_metric_name(
            "by.version",
            artifact_version,
            tf_github_org,
            tf_github_repo,
            deployment_name,
            deployment_type,
            test_name,
            tf_triggering_env,
            metric_name,
            None,
            False,
            build_variant_name,
            running_platform,
        )
        labels = get_project_ts_tags(
            tf_github_org,
            tf_github_repo,
            deployment_name,
            deployment_type,
            tf_triggering_env,
            metadata_tags,
            build_variant_name,
        )
        labels["version"] = artifact_version
        labels["deployment_name+version"] = "{} {}".format(
            deployment_name, artifact_version
        )
        labels["test_name"] = str(test_name)
        labels["metric"] = str(metric_name)
        logging.info(
            "Adding metric {}={} to time-serie named {}".format(
                metric_name, metric_value, tsname_use_case_duration
            )
        )
        ts = {"labels": labels}
        exporter_create_ts(rts, ts, tsname_use_case_duration)
        rts.ts().add(
            tsname_use_case_duration,
            start_time_ms,
            metric_value,
            labels=labels,
        )
    else:
        logging.warning(
            "Given that metric {}={} ( is None ) we will skip adding it to timeseries".format(
                metric_name, metric_value
            )
        )


def timeseries_test_sucess_flow(
    push_results_redistimeseries,
    artifact_version,
    benchmark_config,
    benchmark_duration_seconds,
    dataset_load_duration_seconds,
    default_metrics,
    deployment_name,
    deployment_type,
    exporter_timemetric_path,
    results_dict,
    rts,
    start_time_ms,
    test_name,
    tf_github_branch,
    tf_github_org,
    tf_github_repo,
    tf_triggering_env,
    metadata_tags={},
    build_variant_name=None,
    running_platform=None,
    timeseries_dict=None,
    git_hash=None,
):
    testcase_metric_context_paths = []
    version_target_tables = None
    branch_target_tables = None

    if timeseries_dict is None:
        (
            timeseries_dict,
            testcase_metric_context_paths,
            version_target_tables,
            branch_target_tables,
        ) = prepare_timeseries_dict(
            artifact_version,
            benchmark_config,
            default_metrics,
            deployment_name,
            deployment_type,
            exporter_timemetric_path,
            results_dict,
            test_name,
            tf_github_branch,
            tf_github_org,
            tf_github_repo,
            tf_triggering_env,
            metadata_tags,
            build_variant_name,
            running_platform,
            start_time_ms,
            git_hash,
        )
    if push_results_redistimeseries:
        logging.info(
            "Pushing results to RedisTimeSeries. Have {} distinct data-points to insert.".format(
                len(timeseries_dict.keys())
            )
        )
        push_data_to_redistimeseries(rts, timeseries_dict)
        if version_target_tables is not None:
            logging.info(
                "There are a total of {} distinct target tables by version".format(
                    len(version_target_tables.keys())
                )
            )
            for (
                version_target_table_keyname,
                version_target_table_dict,
            ) in version_target_tables.items():
                logging.info(
                    "Setting target table by version on key {}".format(
                        version_target_table_keyname
                    )
                )
                if "contains-target" in version_target_table_dict:
                    del version_target_table_dict["contains-target"]
                rts.hset(
                    version_target_table_keyname, None, None, version_target_table_dict
                )
        if branch_target_tables is not None:
            logging.info(
                "There are a total of {} distinct target tables by branch".format(
                    len(branch_target_tables.keys())
                )
            )
            for (
                branch_target_table_keyname,
                branch_target_table_dict,
            ) in branch_target_tables.items():

                logging.info(
                    "Setting target table by branch on key {}".format(
                        branch_target_table_keyname
                    )
                )
                if "contains-target" in branch_target_table_dict:
                    del branch_target_table_dict["contains-target"]
                rts.hset(
                    branch_target_table_keyname, None, None, branch_target_table_dict
                )
        if test_name is not None:
            if type(test_name) is str:
                update_secondary_result_keys(
                    artifact_version,
                    benchmark_duration_seconds,
                    build_variant_name,
                    dataset_load_duration_seconds,
                    deployment_name,
                    deployment_type,
                    metadata_tags,
                    rts,
                    running_platform,
                    start_time_ms,
                    test_name,
                    testcase_metric_context_paths,
                    tf_github_branch,
                    tf_github_org,
                    tf_github_repo,
                    tf_triggering_env,
                )
            if type(test_name) is list:
                for inner_test_name in test_name:
                    update_secondary_result_keys(
                        artifact_version,
                        benchmark_duration_seconds,
                        build_variant_name,
                        dataset_load_duration_seconds,
                        deployment_name,
                        deployment_type,
                        metadata_tags,
                        rts,
                        running_platform,
                        start_time_ms,
                        inner_test_name,
                        testcase_metric_context_paths,
                        tf_github_branch,
                        tf_github_org,
                        tf_github_repo,
                        tf_triggering_env,
                    )
        else:
            update_secondary_result_keys(
                artifact_version,
                benchmark_duration_seconds,
                build_variant_name,
                dataset_load_duration_seconds,
                deployment_name,
                deployment_type,
                metadata_tags,
                rts,
                running_platform,
                start_time_ms,
                test_name,
                testcase_metric_context_paths,
                tf_github_branch,
                tf_github_org,
                tf_github_repo,
                tf_triggering_env,
            )
    return version_target_tables, branch_target_tables


def update_secondary_result_keys(
    artifact_version,
    benchmark_duration_seconds,
    build_variant_name,
    dataset_load_duration_seconds,
    deployment_name,
    deployment_type,
    metadata_tags,
    rts,
    running_platform,
    start_time_ms,
    test_name,
    testcase_metric_context_paths,
    tf_github_branch,
    tf_github_org,
    tf_github_repo,
    tf_triggering_env,
):
    (
        _,
        testcases_setname,
        deployment_name_zsetname,
        _,
        tsname_project_total_success,
        running_platforms_setname,
        build_variant_setname,
        testcases_metric_context_path_setname,
        testcases_and_metric_context_path_setname,
        project_archs_setname,
        project_oss_setname,
        project_branches_setname,
        project_versions_setname,
        project_compilers_setname,
    ) = get_overall_dashboard_keynames(
        tf_github_org,
        tf_github_repo,
        tf_triggering_env,
        build_variant_name,
        running_platform,
        test_name,
    )
    try:
        rts.zadd(deployment_name_zsetname, {deployment_name: start_time_ms})
        if test_name is not None:
            deployment_name_zsetname_testnames = (
                deployment_name_zsetname
                + "{}:deployment_name={}".format(
                    deployment_name_zsetname, deployment_name
                )
            )
            rts.zadd(deployment_name_zsetname_testnames, {test_name: start_time_ms})
            rts.sadd(testcases_setname, test_name)
            testcases_zsetname = testcases_setname + ":zset"
            rts.zadd(testcases_zsetname, {test_name: start_time_ms})
            if "component" in metadata_tags:
                testcases_zsetname_component = "{}:zset:component:{}".format(
                    testcases_setname, metadata_tags["component"]
                )
                rts.zadd(testcases_zsetname_component, {test_name: start_time_ms})
        if "arch" in metadata_tags:
            rts.sadd(project_archs_setname, metadata_tags["arch"])
        if "os" in metadata_tags:
            rts.sadd(project_oss_setname, metadata_tags["os"])
        if "compiler" in metadata_tags:
            rts.sadd(project_compilers_setname, metadata_tags["compiler"])
        if tf_github_branch is not None and tf_github_branch != "":
            rts.sadd(project_branches_setname, tf_github_branch)
            project_branches_zsetname = project_branches_setname + ":zset"
            rts.zadd(project_branches_zsetname, {tf_github_branch: start_time_ms})
        if artifact_version is not None and artifact_version != "":
            rts.sadd(project_versions_setname, artifact_version)
            project_versions_zsetname = project_versions_setname + ":zset"
            rts.zadd(project_versions_zsetname, {artifact_version: start_time_ms})
        if running_platform is not None:
            rts.sadd(running_platforms_setname, running_platform)
            running_platforms_szetname = running_platforms_setname + ":zset"
            rts.zadd(running_platforms_szetname, {running_platform: start_time_ms})
        if build_variant_name is not None:
            rts.sadd(build_variant_setname, build_variant_name)
            build_variant_zsetname = build_variant_setname + ":zset"
            rts.zadd(build_variant_zsetname, {build_variant_name: start_time_ms})
        if testcase_metric_context_paths is not None:
            for metric_context_path in testcase_metric_context_paths:
                if testcases_metric_context_path_setname != "":
                    rts.sadd(testcases_metric_context_path_setname, metric_context_path)
                    rts.sadd(
                        testcases_and_metric_context_path_setname,
                        "{}:{}".format(test_name, metric_context_path),
                    )
        rts.ts().incrby(
            tsname_project_total_success,
            1,
            timestamp=start_time_ms,
            labels=get_project_ts_tags(
                tf_github_org,
                tf_github_repo,
                deployment_name,
                deployment_type,
                tf_triggering_env,
                metadata_tags,
                build_variant_name,
                running_platform,
            ),
        )
        if tf_github_branch is not None and tf_github_branch != "":
            add_standardized_metric_bybranch(
                "benchmark_duration",
                benchmark_duration_seconds,
                str(tf_github_branch),
                deployment_name,
                deployment_type,
                rts,
                start_time_ms,
                test_name,
                tf_github_org,
                tf_github_repo,
                tf_triggering_env,
                metadata_tags,
                build_variant_name,
                running_platform,
            )
            add_standardized_metric_bybranch(
                "dataset_load_duration",
                dataset_load_duration_seconds,
                str(tf_github_branch),
                deployment_name,
                deployment_type,
                rts,
                start_time_ms,
                test_name,
                tf_github_org,
                tf_github_repo,
                tf_triggering_env,
                metadata_tags,
                build_variant_name,
                running_platform,
            )
        if artifact_version is not None and artifact_version != "":
            add_standardized_metric_byversion(
                "benchmark_duration",
                benchmark_duration_seconds,
                artifact_version,
                deployment_name,
                deployment_type,
                rts,
                start_time_ms,
                test_name,
                tf_github_org,
                tf_github_repo,
                tf_triggering_env,
                metadata_tags,
                build_variant_name,
                running_platform,
            )
            add_standardized_metric_byversion(
                "dataset_load_duration",
                dataset_load_duration_seconds,
                artifact_version,
                deployment_name,
                deployment_type,
                rts,
                start_time_ms,
                test_name,
                tf_github_org,
                tf_github_repo,
                tf_triggering_env,
                metadata_tags,
                build_variant_name,
                running_platform,
            )
    except redis.exceptions.ResponseError as e:
        logging.warning(
            "Error while updating secondary data structures {}. ".format(e.__str__())
        )
        pass


def get_start_time_vars(start_time=None):
    if start_time is None:
        start_time = datetime.datetime.utcnow()
    start_time_ms = int(
        (start_time - datetime.datetime(1970, 1, 1)).total_seconds() * 1000
    )
    start_time_str = start_time.strftime("%Y-%m-%d-%H-%M-%S")
    return start_time, start_time_ms, start_time_str


def timeseries_test_failure_flow(
    args,
    deployment_name,
    deployment_type,
    rts,
    start_time_ms,
    tf_github_org,
    tf_github_repo,
    tf_triggering_env,
    tsname_project_total_failures,
):
    if args.push_results_redistimeseries:
        if start_time_ms is None:
            _, start_time_ms, _ = get_start_time_vars()
        try:
            rts.ts().incrby(
                tsname_project_total_failures,
                1,
                timestamp=start_time_ms,
                labels=get_project_ts_tags(
                    tf_github_org,
                    tf_github_repo,
                    deployment_name,
                    deployment_type,
                    tf_triggering_env,
                ),
            )
        except redis.exceptions.ResponseError as e:
            logging.warning(
                "Error while updating secondary data structures {}. ".format(
                    e.__str__()
                )
            )
            pass


def datasink_profile_tabular_data(
    github_branch,
    github_org_name,
    github_repo_name,
    github_sha,
    overall_tabular_data_map,
    rts,
    setup_type,
    start_time_ms,
    start_time_str,
    test_name,
    tf_triggering_env,
):
    zset_profiles_key_name = get_profilers_rts_key_prefix(
        tf_triggering_env,
        github_org_name,
        github_repo_name,
    )
    profile_test_suffix = "{start_time_str}:{test_name}/{setup_type}/{github_branch}/{github_hash}".format(
        start_time_str=start_time_str,
        test_name=test_name,
        setup_type=setup_type,
        github_branch=github_branch,
        github_hash=github_sha,
    )
    rts.zadd(
        zset_profiles_key_name,
        {profile_test_suffix: start_time_ms},
    )
    for (
        profile_tabular_type,
        tabular_data,
    ) in overall_tabular_data_map.items():
        tabular_suffix = "{}:{}".format(profile_tabular_type, profile_test_suffix)
        logging.info(
            "Pushing to data-sink tabular data from pprof ({}). Tabular suffix: {}".format(
                profile_tabular_type, tabular_suffix
            )
        )

        table_columns_text_key = "{}:columns:text".format(tabular_suffix)
        table_columns_type_key = "{}:columns:type".format(tabular_suffix)
        logging.info(
            "Pushing list key (named {}) the following column text: {}".format(
                table_columns_text_key, tabular_data["columns:text"]
            )
        )
        rts.rpush(table_columns_text_key, *tabular_data["columns:text"])
        logging.info(
            "Pushing list key (named {}) the following column types: {}".format(
                table_columns_type_key, tabular_data["columns:type"]
            )
        )
        rts.rpush(table_columns_type_key, *tabular_data["columns:type"])
        for row_name in tabular_data["columns:text"]:
            table_row_key = "{}:rows:{}".format(tabular_suffix, row_name)
            row_values = tabular_data["rows:{}".format(row_name)]
            rts.rpush(table_row_key, *row_values)



import os
import yaml


def get_topologies(usecase_filename):
    full_filename = os.path.abspath(usecase_filename)
    topologies_map = {}
    with open(full_filename) as stream:
        build_config = yaml.safe_load(stream)
        spec = build_config["spec"]
        setups = spec["setups"]
        for topology in setups:
            topology_name = topology["name"]
            topologies_map[topology_name] = topology
        # print(build_config)

    return topologies_map



import math


def generate_cpuset_cpus(ceil_db_cpu_limit, current_cpu_pos):
    previous_cpu_pos = current_cpu_pos
    current_cpu_pos = current_cpu_pos + int(ceil_db_cpu_limit)
    db_cpuset_cpus = ",".join(
        [str(x) for x in range(previous_cpu_pos, current_cpu_pos)]
    )
    return db_cpuset_cpus, current_cpu_pos


def extract_db_cpu_limit(topologies_map, topology_spec_name):
    topology_spec = topologies_map[topology_spec_name]
    db_cpu_limit = topology_spec["resources"]["requests"]["cpus"]
    ceil_db_cpu_limit = math.ceil(float(db_cpu_limit))
    return ceil_db_cpu_limit



import datetime
import json
import logging
import pathlib
import shutil
import tempfile
import traceback
import re
import docker
import redis
import os
from pathlib import Path
import sys
import time

from docker.models.containers import Container
from redisbench_admin.profilers.profilers_local import (
    check_compatible_system_and_kernel_and_prepare_profile,
)

from redis_benchmarks_specification.__common__.env import (
    LOG_FORMAT,
    LOG_DATEFMT,
    LOG_LEVEL,
    REDIS_HEALTH_CHECK_INTERVAL,
    REDIS_SOCKET_TIMEOUT,
    REDIS_BINS_EXPIRE_SECS,
)
from redis_benchmarks_specification.__common__.github import (
    check_github_available_and_actionable,
    check_benchmark_running_comment,
    update_comment_if_needed,
    create_new_pr_comment,
    generate_benchmark_started_pr_comment,
    check_regression_comment,
)
from redis_benchmarks_specification.__common__.package import (
    get_version_string,
    populate_with_poetry_data,
)
from redis_benchmarks_specification.__common__.runner import (
    extract_testsuites,
    reset_commandstats,
    exporter_datasink_common,
    execute_init_commands,
)
from redis_benchmarks_specification.__common__.timeseries import (
    datasink_profile_tabular_data,
)
from redis_benchmarks_specification.__compare__.compare import (
    compute_regression_table,
    prepare_regression_comment,
    extract_default_branch_and_metric,
)
from redis_benchmarks_specification.__runner__.runner import (
    print_results_table_stdout,
    prepare_memtier_benchmark_parameters,
)
from redis_benchmarks_specification.__self_contained_coordinator__.args import (
    create_self_contained_coordinator_args,
)
from redis_benchmarks_specification.__self_contained_coordinator__.runners import (
    build_runners_consumer_group_create,
    get_runners_consumer_group_name,
)
from redis_benchmarks_specification.__setups__.topologies import get_topologies


from redisbench_admin.profilers.profilers_local import (
    local_profilers_platform_checks,
    profilers_start_if_required,
    profilers_stop_if_required,
)
from redisbench_admin.run.common import (
    get_start_time_vars,
    prepare_benchmark_parameters,
)
from redisbench_admin.run.grafana import generate_artifacts_table_grafana_redis

from redisbench_admin.run.run import calculate_client_tool_duration_and_check
from redisbench_admin.utils.benchmark_config import (
    get_final_benchmark_config,
    get_defaults,
)
from redisbench_admin.utils.local import get_local_run_full_filename
from redisbench_admin.utils.results import post_process_benchmark_results

from redis_benchmarks_specification.__common__.env import (
    STREAM_KEYNAME_NEW_BUILD_EVENTS,
    S3_BUCKET_NAME,
)
from redis_benchmarks_specification.__common__.spec import (
    extract_build_variant_variations,
    extract_client_cpu_limit,
    extract_client_tool,
    extract_client_container_image,
    extract_redis_dbconfig_parameters,
    extract_redis_configuration_from_topology,
)
from redis_benchmarks_specification.__self_contained_coordinator__.artifacts import (
    restore_build_artifacts_from_test_details,
)
from redis_benchmarks_specification.__self_contained_coordinator__.build_info import (
    extract_build_info_from_streamdata,
)
from redis_benchmarks_specification.__self_contained_coordinator__.cpuset import (
    extract_db_cpu_limit,
    generate_cpuset_cpus,
)
from redis_benchmarks_specification.__self_contained_coordinator__.docker import (
    generate_standalone_redis_server_args,
)


def main():
    _, _, project_version = populate_with_poetry_data()
    project_name = "redis-benchmarks-spec runner(self-contained)"
    parser = create_self_contained_coordinator_args(
        get_version_string(project_name, project_version)
    )
    args = parser.parse_args()

    if args.logname is not None:
        print("Writting log to {}".format(args.logname))
        logging.basicConfig(
            filename=args.logname,
            filemode="a",
            format=LOG_FORMAT,
            datefmt=LOG_DATEFMT,
            level=LOG_LEVEL,
        )
    else:
        # logging settings
        logging.basicConfig(
            format=LOG_FORMAT,
            level=LOG_LEVEL,
            datefmt=LOG_DATEFMT,
        )
    logging.info(get_version_string(project_name, project_version))
    topologies_folder = os.path.abspath(args.setups_folder + "/topologies")
    logging.info("Using topologies folder dir {}".format(topologies_folder))
    topologies_files = pathlib.Path(topologies_folder).glob("*.yml")
    topologies_files = [str(x) for x in topologies_files]
    logging.info(
        "Reading topologies specifications from: {}".format(
            " ".join([str(x) for x in topologies_files])
        )
    )
    topologies_map = get_topologies(topologies_files[0])
    testsuite_spec_files = extract_testsuites(args)

    logging.info(
        "Reading event streams from: {}:{} with user {}".format(
            args.event_stream_host, args.event_stream_port, args.event_stream_user
        )
    )
    try:
        gh_event_conn = redis.StrictRedis(
            host=args.event_stream_host,
            port=args.event_stream_port,
            decode_responses=False,  # dont decode due to binary archives
            password=args.event_stream_pass,
            username=args.event_stream_user,
            health_check_interval=REDIS_HEALTH_CHECK_INTERVAL,
            socket_connect_timeout=REDIS_SOCKET_TIMEOUT,
            socket_keepalive=True,
        )
        gh_event_conn.ping()
    except redis.exceptions.ConnectionError as e:
        logging.error(
            "Unable to connect to redis available at: {}:{} to read the event streams".format(
                args.event_stream_host, args.event_stream_port
            )
        )
        logging.error("Error message {}".format(e.__str__()))
        exit(1)
    datasink_conn = None
    if args.datasink_push_results_redistimeseries:
        logging.info(
            "Checking redistimeseries datasink connection is available at: {}:{} to push the timeseries data".format(
                args.datasink_redistimeseries_host, args.datasink_redistimeseries_port
            )
        )
        try:
            datasink_conn = redis.StrictRedis(
                host=args.datasink_redistimeseries_host,
                port=args.datasink_redistimeseries_port,
                decode_responses=True,
                password=args.datasink_redistimeseries_pass,
                username=args.datasink_redistimeseries_user,
                health_check_interval=REDIS_HEALTH_CHECK_INTERVAL,
                socket_connect_timeout=REDIS_SOCKET_TIMEOUT,
                socket_keepalive=True,
            )
            datasink_conn.ping()
        except redis.exceptions.ConnectionError as e:
            logging.error(
                "Unable to connect to redis available at: {}:{}".format(
                    args.datasink_redistimeseries_host,
                    args.datasink_redistimeseries_port,
                )
            )
            logging.error("Error message {}".format(e.__str__()))
            exit(1)

    logging.info("checking build spec requirements")
    running_platform = args.platform_name
    build_runners_consumer_group_create(gh_event_conn, running_platform)
    stream_id = None
    docker_client = docker.from_env()
    home = str(Path.home())
    cpuset_start_pos = args.cpuset_start_pos
    logging.info("Start CPU pinning at position {}".format(cpuset_start_pos))
    redis_proc_start_port = args.redis_proc_start_port
    logging.info("Redis Processes start port: {}".format(redis_proc_start_port))

    priority_lower_limit = args.tests_priority_lower_limit
    priority_upper_limit = args.tests_priority_upper_limit

    logging.info(
        f"Using priority for test filters [{priority_lower_limit},{priority_upper_limit}]"
    )

    default_baseline_branch, default_metrics_str = extract_default_branch_and_metric(
        args.defaults_filename
    )

    # TODO: confirm we do have enough cores to run the spec
    # availabe_cpus = args.cpu_count
    datasink_push_results_redistimeseries = args.datasink_push_results_redistimeseries
    grafana_profile_dashboard = args.grafana_profile_dashboard

    defaults_filename = args.defaults_filename
    (
        _,
        _,
        default_metrics,
        _,
        _,
        _,
    ) = get_defaults(defaults_filename)

    # Consumer id
    consumer_pos = args.consumer_pos
    logging.info("Consumer pos {}".format(consumer_pos))

    # Arch
    arch = args.arch
    logging.info("Running for arch: {}".format(arch))

    # Github token
    github_token = args.github_token
    if github_token is not None:
        logging.info("Detected GITHUB token. will push PR comments with updates")

    # Docker air gap usage
    docker_air_gap = args.docker_air_gap
    if docker_air_gap:
        logging.info(
            "Using docker in an air-gapped way. Restoring running images from redis keys."
        )

    profilers_list = []
    profilers_enabled = args.enable_profilers
    if profilers_enabled:
        profilers_list = args.profilers.split(",")
        res = check_compatible_system_and_kernel_and_prepare_profile(args)
        if res is False:
            logging.error(
                "Requested for the following profilers to be enabled but something went wrong: {}.".format(
                    " ".join(profilers_list)
                )
            )
            exit(1)

    override_memtier_test_time = args.override_memtier_test_time
    if override_memtier_test_time > 0:
        logging.info(
            "Overriding memtier benchmark --test-time to {} seconds".format(
                override_memtier_test_time
            )
        )
    logging.info("Entering blocking read waiting for work.")
    if stream_id is None:
        stream_id = args.consumer_start_id
    while True:
        _, stream_id, _, _ = self_contained_coordinator_blocking_read(
            gh_event_conn,
            datasink_push_results_redistimeseries,
            docker_client,
            home,
            stream_id,
            datasink_conn,
            testsuite_spec_files,
            topologies_map,
            running_platform,
            profilers_enabled,
            profilers_list,
            grafana_profile_dashboard,
            cpuset_start_pos,
            redis_proc_start_port,
            consumer_pos,
            docker_air_gap,
            override_memtier_test_time,
            default_metrics,
            arch,
            github_token,
            priority_lower_limit,
            priority_upper_limit,
            default_baseline_branch,
            default_metrics_str,
        )


def check_health(container):
    logging.info(container.attrs["State"])
    health_status = container.attrs["State"].get("Health", {}).get("Status")
    return health_status


def self_contained_coordinator_blocking_read(
    github_event_conn,
    datasink_push_results_redistimeseries,
    docker_client,
    home,
    stream_id,
    datasink_conn,
    testsuite_spec_files,
    topologies_map,
    platform_name,
    profilers_enabled,
    profilers_list,
    grafana_profile_dashboard="",
    cpuset_start_pos=0,
    redis_proc_start_port=6379,
    consumer_pos=1,
    docker_air_gap=False,
    override_test_time=1,
    default_metrics=None,
    arch="amd64",
    github_token=None,
    priority_lower_limit=0,
    priority_upper_limit=10000,
    default_baseline_branch="unstable",
    default_metrics_str="ALL_STATS.Totals.Ops/sec",
    docker_keep_env=False,
    restore_build_artifacts_default=True,
):
    num_process_streams = 0
    num_process_test_suites = 0
    overall_result = False
    consumer_name = "{}-self-contained-proc#{}".format(
        get_runners_consumer_group_name(platform_name), consumer_pos
    )
    logging.info(
        "Consuming from group {}. Consumer id {}".format(
            get_runners_consumer_group_name(platform_name), consumer_name
        )
    )
    newTestInfo = github_event_conn.xreadgroup(
        get_runners_consumer_group_name(platform_name),
        consumer_name,
        {STREAM_KEYNAME_NEW_BUILD_EVENTS: stream_id},
        count=1,
        block=0,
    )
    if len(newTestInfo[0]) < 2 or len(newTestInfo[0][1]) < 1:
        stream_id = ">"
    else:
        (
            stream_id,
            overall_result,
            total_test_suite_runs,
        ) = process_self_contained_coordinator_stream(
            github_event_conn,
            datasink_push_results_redistimeseries,
            docker_client,
            home,
            newTestInfo,
            datasink_conn,
            testsuite_spec_files,
            topologies_map,
            platform_name,
            profilers_enabled,
            profilers_list,
            grafana_profile_dashboard,
            cpuset_start_pos,
            redis_proc_start_port,
            docker_air_gap,
            "defaults.yml",
            override_test_time,
            default_metrics,
            arch,
            github_token,
            priority_lower_limit,
            priority_upper_limit,
            default_baseline_branch,
            default_metrics_str,
            docker_keep_env,
            restore_build_artifacts_default,
        )
        num_process_streams = num_process_streams + 1
        num_process_test_suites = num_process_test_suites + total_test_suite_runs
        if overall_result is True:
            ack_reply = github_event_conn.xack(
                STREAM_KEYNAME_NEW_BUILD_EVENTS,
                get_runners_consumer_group_name(platform_name),
                stream_id,
            )
            if type(ack_reply) == bytes:
                ack_reply = ack_reply.decode()
            if ack_reply == "1" or ack_reply == 1:
                logging.info(
                    "Sucessfully acknowledge BENCHMARK variation stream with id {}.".format(
                        stream_id
                    )
                )
            else:
                logging.error(
                    "Unable to acknowledge build variation stream with id {}. XACK reply {}".format(
                        stream_id, ack_reply
                    )
                )
    return overall_result, stream_id, num_process_streams, num_process_test_suites


#
# def prepare_memtier_benchmark_parameters(
#     clientconfig,
#     full_benchmark_path,
#     port,
#     server,
#     local_benchmark_output_filename,
#     oss_cluster_api_enabled,
# ):
#     benchmark_command = [
#         full_benchmark_path,
#         "--port",
#         "{}".format(port),
#         "--server",
#         "{}".format(server),
#         "--json-out-file",
#         local_benchmark_output_filename,
#     ]
#     if oss_cluster_api_enabled is True:
#         benchmark_command.append("--cluster-mode")
#     benchmark_command_str = " ".join(benchmark_command)
#     if "arguments" in clientconfig:
#         benchmark_command_str = benchmark_command_str + " " + clientconfig["arguments"]
#
#     return None, benchmark_command_str


def process_self_contained_coordinator_stream(
    github_event_conn,
    datasink_push_results_redistimeseries,
    docker_client,
    home,
    newTestInfo,
    datasink_conn,
    testsuite_spec_files,
    topologies_map,
    running_platform,
    profilers_enabled=False,
    profilers_list=[],
    grafana_profile_dashboard="",
    cpuset_start_pos=0,
    redis_proc_start_port=6379,
    default_docker_air_gap=False,
    defaults_filename="defaults.yml",
    override_test_time=0,
    default_metrics=[],
    arch="amd64",
    github_token=None,
    priority_lower_limit=0,
    priority_upper_limit=10000,
    default_baseline_branch="unstable",
    default_metrics_str="ALL_STATS.Totals.Ops/sec",
    docker_keep_env=False,
    restore_build_artifacts_default=True,
):
    stream_id = "n/a"
    overall_result = False
    total_test_suite_runs = 0
    # github updates
    is_actionable_pr = False
    contains_benchmark_run_comment = False
    github_pr = None
    old_benchmark_run_comment_body = ""
    pr_link = ""
    regression_comment = None
    pull_request = None
    auto_approve_github = True
    # defaults
    default_github_org = "redis"
    default_github_repo = "redis"
    restore_build_artifacts = restore_build_artifacts_default

    try:
        stream_id, testDetails = newTestInfo[0][1][0]
        stream_id = stream_id.decode()
        logging.info("Received work . Stream id {}.".format(stream_id))

        if b"run_image" in testDetails:
            (
                build_variant_name,
                metadata,
                build_artifacts,
                git_hash,
                git_branch,
                git_version,
                run_image,
                use_git_timestamp,
                git_timestamp_ms,
                run_arch,
            ) = extract_build_info_from_streamdata(testDetails)
            tf_github_org = default_github_org
            if b"github_org" in testDetails:
                tf_github_org = testDetails[b"github_org"].decode()
                logging.info(
                    f"detected a github_org definition on the streamdata: {tf_github_org}. Overriding the default one: {default_github_org}"
                )
            tf_github_repo = default_github_repo
            if b"github_repo" in testDetails:
                tf_github_repo = testDetails[b"github_repo"].decode()
                logging.info(
                    f"detected a github_org definition on the streamdata: {tf_github_repo}. Overriding the default one: {default_github_repo}"
                )

            mnt_point = "/mnt/redis/"
            if b"mnt_point" in testDetails:
                mnt_point = testDetails[b"mnt_point"].decode()
                logging.info(
                    f"detected a mnt_point definition on the streamdata: {mnt_point}."
                )

            executable = f"{mnt_point}redis-server"
            if b"executable" in testDetails:
                executable = testDetails[b"executable"].decode()
                logging.info(
                    f"detected a executable definition on the streamdata: {executable}."
                )

            server_name = "redis"
            if b"server_name" in testDetails:
                server_name = testDetails[b"server_name"].decode()
                logging.info(
                    f"detected a server_name definition on the streamdata: {server_name}."
                )
                new_executable = f"{mnt_point}{server_name}-server"
                logging.info(
                    "changing executable from {executable} to {new_executable}"
                )
                executable = new_executable

            if b"restore_build_artifacts" in testDetails:
                restore_build_artifacts = bool(
                    testDetails[b"restore_build_artifacts"].decode()
                )
                logging.info(
                    f"detected a restore_build_artifacts config {restore_build_artifacts} overriding the default just for this test"
                )

            test_docker_air_gap = default_docker_air_gap
            # check if we override the docker air gap on this test details
            if b"docker_air_gap" in testDetails:
                test_docker_air_gap = bool(testDetails[b"docker_air_gap"].decode())
                logging.info(
                    f"detected a docker air gap config {test_docker_air_gap} overriding the default of {default_docker_air_gap} just for this test"
                )

            if b"priority_upper_limit" in testDetails:
                stream_priority_upper_limit = int(
                    testDetails[b"priority_upper_limit"].decode()
                )
                logging.info(
                    f"detected a priority_upper_limit definition on the streamdata {stream_priority_upper_limit}. will replace the default upper limit of {priority_upper_limit}"
                )
                priority_upper_limit = stream_priority_upper_limit

            if b"priority_lower_limit" in testDetails:
                stream_priority_lower_limit = int(
                    testDetails[b"priority_lower_limit"].decode()
                )
                logging.info(
                    f"detected a priority_lower_limit definition on the streamdata {stream_priority_lower_limit}. will replace the default lower limit of {priority_lower_limit}"
                )
                priority_lower_limit = stream_priority_lower_limit

            if b"pull_request" in testDetails:
                pull_request = testDetails[b"pull_request"].decode()
                logging.info(
                    f"detected a pull_request definition on the streamdata {pull_request}"
                )
                verbose = True
                fn = check_benchmark_running_comment
                (
                    contains_benchmark_run_comment,
                    github_pr,
                    is_actionable_pr,
                    old_benchmark_run_comment_body,
                    pr_link,
                    benchmark_run_comment,
                ) = check_github_available_and_actionable(
                    fn, github_token, pull_request, "redis", "redis", verbose
                )

            tests_regexp = ".*"
            if b"tests_regexp" in testDetails:
                tests_regexp = testDetails[b"tests_regexp"].decode()
                logging.info(
                    f"detected a regexp definition on the streamdata {tests_regexp}"
                )

            command_groups_regexp = None
            if b"tests_groups_regexp" in testDetails:
                command_groups_regexp = testDetails[b"tests_groups_regexp"].decode()
                logging.info(
                    f"detected a command groups regexp definition on the streamdata {command_groups_regexp}"
                )

            skip_test = False
            if b"platform" in testDetails:
                platform = testDetails[b"platform"]
                if running_platform != platform:
                    skip_test = True
                    logging.info(
                        "skipping stream_id {} given plaform {}!={}".format(
                            stream_id, running_platform, platform
                        )
                    )

            if run_arch != arch:
                skip_test = True
                logging.info(
                    "skipping stream_id {} given arch {}!={}".format(
                        stream_id, run_arch, arch
                    )
                )

            if skip_test is False:
                overall_result = True
                profiler_dashboard_links = []
                if test_docker_air_gap:
                    airgap_key = "docker:air-gap:{}".format(run_image)
                    logging.info(
                        "Restoring docker image: {} from {}".format(
                            run_image, airgap_key
                        )
                    )
                    airgap_docker_image_bin = github_event_conn.get(airgap_key)
                    images_loaded = docker_client.images.load(airgap_docker_image_bin)
                    logging.info("Successfully loaded images {}".format(images_loaded))

                stream_time_ms = stream_id.split("-")[0]
                zset_running_platform_benchmarks = f"ci.benchmarks.redis/ci/redis/redis:benchmarks:{running_platform}:zset"
                res = github_event_conn.zadd(
                    zset_running_platform_benchmarks,
                    {stream_id: stream_time_ms},
                )
                logging.info(
                    f"Added stream with id {stream_id} to zset {zset_running_platform_benchmarks}. res={res}"
                )

                stream_test_list_pending = f"ci.benchmarks.redis/ci/redis/redis:benchmarks:{stream_id}:{running_platform}:tests_pending"
                stream_test_list_running = f"ci.benchmarks.redis/ci/redis/redis:benchmarks:{stream_id}:{running_platform}:tests_running"
                stream_test_list_failed = f"ci.benchmarks.redis/ci/redis/redis:benchmarks:{stream_id}:{running_platform}:tests_failed"
                stream_test_list_completed = f"ci.benchmarks.redis/ci/redis/redis:benchmarks:{stream_id}:{running_platform}:tests_completed"

                filtered_test_files = filter_test_files(
                    defaults_filename,
                    priority_lower_limit,
                    priority_upper_limit,
                    tests_regexp,
                    testsuite_spec_files,
                    command_groups_regexp,
                )

                for test_file in filtered_test_files:
                    with open(test_file, "r") as stream:
                        (
                            _,
                            benchmark_config,
                            test_name,
                        ) = get_final_benchmark_config(None, None, stream, "")
                    github_event_conn.lpush(stream_test_list_pending, test_name)
                    github_event_conn.expire(
                        stream_test_list_pending, REDIS_BINS_EXPIRE_SECS
                    )
                    logging.info(
                        f"Added test named {test_name} to the pending test list in key {stream_test_list_pending}"
                    )

                pending_tests = len(filtered_test_files)
                failed_tests = 0
                benchmark_suite_start_datetime = datetime.datetime.utcnow()
                # update on github if needed
                if is_actionable_pr:
                    comment_body = generate_benchmark_started_pr_comment(
                        stream_id,
                        pending_tests,
                        len(filtered_test_files),
                        failed_tests,
                        benchmark_suite_start_datetime,
                        0,
                    )
                    if contains_benchmark_run_comment:
                        update_comment_if_needed(
                            auto_approve_github,
                            comment_body,
                            old_benchmark_run_comment_body,
                            benchmark_run_comment,
                            verbose,
                        )
                    else:
                        benchmark_run_comment = create_new_pr_comment(
                            auto_approve_github, comment_body, github_pr, pr_link
                        )

                for test_file in filtered_test_files:
                    redis_containers = []
                    client_containers = []
                    with open(test_file, "r") as stream:
                        (
                            _,
                            benchmark_config,
                            test_name,
                        ) = get_final_benchmark_config(None, None, stream, "")
                        github_event_conn.lrem(stream_test_list_pending, 1, test_name)
                        github_event_conn.lpush(stream_test_list_running, test_name)
                        github_event_conn.expire(
                            stream_test_list_running, REDIS_BINS_EXPIRE_SECS
                        )
                        logging.info(
                            f"Added test named {test_name} to the pending test list in key {stream_test_list_running}"
                        )
                        (
                            _,
                            _,
                            redis_configuration_parameters,
                            _,
                            _,
                        ) = extract_redis_dbconfig_parameters(
                            benchmark_config, "dbconfig"
                        )
                        build_variants = extract_build_variant_variations(
                            benchmark_config
                        )
                        if build_variants is not None:
                            logging.info("Detected build variant filter")
                            if build_variant_name not in build_variants:
                                logging.info(
                                    "Skipping {} given it's not part of build-variants for this test-suite {}".format(
                                        build_variant_name, build_variants
                                    )
                                )
                                continue
                            else:
                                logging.info(
                                    "Running build variant {} given it's present on the build-variants spec {}".format(
                                        build_variant_name, build_variants
                                    )
                                )
                        for topology_spec_name in benchmark_config["redis-topologies"]:
                            setup_name = topology_spec_name
                            setup_type = "oss-standalone"
                            if topology_spec_name in topologies_map:
                                topology_spec = topologies_map[topology_spec_name]
                                setup_type = topology_spec["type"]
                            logging.info(
                                f"Running topology named {topology_spec_name} of type {setup_type}"
                            )
                            test_result = False
                            redis_container = None
                            try:
                                current_cpu_pos = cpuset_start_pos
                                ceil_db_cpu_limit = extract_db_cpu_limit(
                                    topologies_map, topology_spec_name
                                )
                                redis_arguments = (
                                    extract_redis_configuration_from_topology(
                                        topologies_map, topology_spec_name
                                    )
                                )
                                temporary_dir = tempfile.mkdtemp(dir=home)
                                temporary_dir_client = tempfile.mkdtemp(dir=home)
                                logging.info(
                                    "Using local temporary dir to persist redis build artifacts. Path: {}".format(
                                        temporary_dir
                                    )
                                )
                                logging.info(
                                    "Using local temporary dir to persist client output files. Path: {}".format(
                                        temporary_dir_client
                                    )
                                )

                                tf_triggering_env = "ci"
                                github_actor = "{}-{}".format(
                                    tf_triggering_env, running_platform
                                )
                                dso = server_name
                                profilers_artifacts_matrix = []

                                collection_summary_str = ""
                                if profilers_enabled:
                                    collection_summary_str = (
                                        local_profilers_platform_checks(
                                            dso,
                                            github_actor,
                                            git_branch,
                                            tf_github_repo,
                                            git_hash,
                                        )
                                    )
                                    logging.info(
                                        "Using the following collection summary string for profiler description: {}".format(
                                            collection_summary_str
                                        )
                                    )
                                if restore_build_artifacts:
                                    restore_build_artifacts_from_test_details(
                                        build_artifacts,
                                        github_event_conn,
                                        temporary_dir,
                                        testDetails,
                                    )

                                command = generate_standalone_redis_server_args(
                                    executable,
                                    redis_proc_start_port,
                                    mnt_point,
                                    redis_configuration_parameters,
                                    redis_arguments,
                                )
                                command_str = " ".join(command)
                                db_cpuset_cpus, current_cpu_pos = generate_cpuset_cpus(
                                    ceil_db_cpu_limit, current_cpu_pos
                                )
                                redis_container = start_redis_container(
                                    command_str,
                                    db_cpuset_cpus,
                                    docker_client,
                                    mnt_point,
                                    redis_containers,
                                    run_image,
                                    temporary_dir,
                                )

                                r = redis.StrictRedis(port=redis_proc_start_port)
                                r.ping()
                                redis_conns = [r]
                                reset_commandstats(redis_conns)
                                redis_pids = []
                                redis_info = r.info()
                                first_redis_pid = redis_info["process_id"]
                                if git_hash is None and "redis_git_sha1" in redis_info:
                                    git_hash = redis_info["redis_git_sha1"]
                                    if (
                                        git_hash == "" or git_hash == 0
                                    ) and "redis_build_id" in redis_info:
                                        git_hash = redis_info["redis_build_id"]
                                    logging.info(
                                        f"Given git_hash was None, we've collected that info from the server reply. git_hash={git_hash}"
                                    )

                                server_version_keyname = f"{server_name}_version"
                                if (
                                    git_version is None
                                    and server_version_keyname in redis_info
                                ):
                                    git_version = redis_info[server_version_keyname]
                                    logging.info(
                                        f"Given git_version was None, we've collected that info from the server reply key named {server_version_keyname}. git_version={git_version}"
                                    )
                                redis_pids.append(first_redis_pid)
                                ceil_client_cpu_limit = extract_client_cpu_limit(
                                    benchmark_config
                                )
                                (
                                    client_cpuset_cpus,
                                    current_cpu_pos,
                                ) = generate_cpuset_cpus(
                                    ceil_client_cpu_limit, current_cpu_pos
                                )
                                client_mnt_point = "/mnt/client/"
                                benchmark_tool_workdir = client_mnt_point

                                if "preload_tool" in benchmark_config["dbconfig"]:
                                    data_prepopulation_step(
                                        benchmark_config,
                                        benchmark_tool_workdir,
                                        client_cpuset_cpus,
                                        docker_client,
                                        git_hash,
                                        redis_proc_start_port,
                                        temporary_dir,
                                        test_name,
                                    )

                                execute_init_commands(
                                    benchmark_config, r, dbconfig_keyname="dbconfig"
                                )

                                benchmark_tool = extract_client_tool(benchmark_config)
                                # backwards compatible
                                if benchmark_tool is None:
                                    benchmark_tool = "redis-benchmark"
                                full_benchmark_path = "/usr/local/bin/{}".format(
                                    benchmark_tool
                                )

                                # setup the benchmark
                                (
                                    start_time,
                                    start_time_ms,
                                    start_time_str,
                                ) = get_start_time_vars()
                                local_benchmark_output_filename = (
                                    get_local_run_full_filename(
                                        start_time_str,
                                        git_hash,
                                        test_name,
                                        setup_name,
                                    )
                                )
                                logging.info(
                                    "Will store benchmark json output to local file {}".format(
                                        local_benchmark_output_filename
                                    )
                                )
                                if "memtier_benchmark" not in benchmark_tool:
                                    # prepare the benchmark command
                                    (
                                        benchmark_command,
                                        benchmark_command_str,
                                    ) = prepare_benchmark_parameters(
                                        benchmark_config,
                                        full_benchmark_path,
                                        redis_proc_start_port,
                                        "localhost",
                                        local_benchmark_output_filename,
                                        False,
                                        benchmark_tool_workdir,
                                        False,
                                    )
                                else:
                                    (
                                        _,
                                        benchmark_command_str,
                                        arbitrary_command,
                                    ) = prepare_memtier_benchmark_parameters(
                                        benchmark_config["clientconfig"],
                                        full_benchmark_path,
                                        redis_proc_start_port,
                                        "localhost",
                                        None,
                                        local_benchmark_output_filename,
                                        False,
                                        False,
                                        False,
                                        None,
                                        None,
                                        None,
                                        None,
                                        override_test_time,
                                    )

                                client_container_image = extract_client_container_image(
                                    benchmark_config
                                )
                                profiler_call_graph_mode = "dwarf"
                                profiler_frequency = 99
                                # start the profile
                                (
                                    profiler_name,
                                    profilers_map,
                                ) = profilers_start_if_required(
                                    profilers_enabled,
                                    profilers_list,
                                    redis_pids,
                                    setup_name,
                                    start_time_str,
                                    test_name,
                                    profiler_frequency,
                                    profiler_call_graph_mode,
                                )

                                logging.info(
                                    "Using docker image {} as benchmark client image (cpuset={}) with the following args: {}".format(
                                        client_container_image,
                                        client_cpuset_cpus,
                                        benchmark_command_str,
                                    )
                                )
                                # run the benchmark
                                benchmark_start_time = datetime.datetime.now()

                                client_container_stdout = docker_client.containers.run(
                                    image=client_container_image,
                                    volumes={
                                        temporary_dir_client: {
                                            "bind": client_mnt_point,
                                            "mode": "rw",
                                        },
                                    },
                                    auto_remove=True,
                                    privileged=True,
                                    working_dir=benchmark_tool_workdir,
                                    command=benchmark_command_str,
                                    network_mode="host",
                                    detach=False,
                                    cpuset_cpus=client_cpuset_cpus,
                                )

                                benchmark_end_time = datetime.datetime.now()
                                benchmark_duration_seconds = (
                                    calculate_client_tool_duration_and_check(
                                        benchmark_end_time, benchmark_start_time
                                    )
                                )
                                logging.info(
                                    "output {}".format(client_container_stdout)
                                )

                                (
                                    _,
                                    overall_tabular_data_map,
                                ) = profilers_stop_if_required(
                                    datasink_push_results_redistimeseries,
                                    benchmark_duration_seconds,
                                    collection_summary_str,
                                    dso,
                                    tf_github_org,
                                    tf_github_repo,
                                    profiler_name,
                                    profilers_artifacts_matrix,
                                    profilers_enabled,
                                    profilers_map,
                                    redis_pids,
                                    S3_BUCKET_NAME,
                                    test_name,
                                )
                                if (
                                    profilers_enabled
                                    and datasink_push_results_redistimeseries
                                ):
                                    datasink_profile_tabular_data(
                                        git_branch,
                                        tf_github_org,
                                        tf_github_repo,
                                        git_hash,
                                        overall_tabular_data_map,
                                        github_event_conn,
                                        setup_name,
                                        start_time_ms,
                                        start_time_str,
                                        test_name,
                                        tf_triggering_env,
                                    )
                                    if len(profilers_artifacts_matrix) == 0:
                                        logging.error(
                                            "No profiler artifact was retrieved"
                                        )
                                    else:
                                        profilers_artifacts = []
                                        for line in profilers_artifacts_matrix:
                                            artifact_name = line[2]
                                            s3_link = line[4]
                                            profilers_artifacts.append(
                                                {
                                                    "artifact_name": artifact_name,
                                                    "s3_link": s3_link,
                                                }
                                            )
                                        https_link = (
                                            generate_artifacts_table_grafana_redis(
                                                datasink_push_results_redistimeseries,
                                                grafana_profile_dashboard,
                                                profilers_artifacts,
                                                datasink_conn,
                                                setup_name,
                                                start_time_ms,
                                                start_time_str,
                                                test_name,
                                                tf_github_org,
                                                tf_github_repo,
                                                git_hash,
                                                git_branch,
                                            )
                                        )
                                        profiler_dashboard_links.append(
                                            [
                                                setup_name,
                                                test_name,
                                                " {} ".format(https_link),
                                            ]
                                        )
                                        logging.info(
                                            "Published new profile info for this testcase. Access it via: {}".format(
                                                https_link
                                            )
                                        )

                                        # Delete all the perf artifacts, now that they are uploaded to S3.
                                        # The .script and .script.mainthread files are not part of the artifacts_matrix and thus have to be deleted separately
                                        line = profilers_artifacts_matrix[0]
                                        logging.info(
                                            "Deleting perf file {}".format(
                                                line[3].split(".")[0]
                                                + ".out.script.mainthread"
                                            )
                                        )
                                        os.remove(
                                            line[3].split(".")[0]
                                            + ".out.script.mainthread"
                                        )
                                        logging.info(
                                            "Deleteing perf file {}".format(
                                                line[3].split(".")[0] + ".out.script"
                                            )
                                        )
                                        os.remove(line[3].split(".")[0] + ".out.script")
                                        for line in profilers_artifacts_matrix:
                                            logging.info(
                                                "Deleting perf file {}".format(line[3])
                                            )
                                            os.remove(line[3])

                                datapoint_time_ms = start_time_ms
                                if (
                                    use_git_timestamp is True
                                    and git_timestamp_ms is not None
                                ):
                                    datapoint_time_ms = git_timestamp_ms
                                post_process_benchmark_results(
                                    benchmark_tool,
                                    local_benchmark_output_filename,
                                    datapoint_time_ms,
                                    start_time_str,
                                    client_container_stdout,
                                    None,
                                )
                                full_result_path = local_benchmark_output_filename
                                if "memtier_benchmark" in benchmark_tool:
                                    full_result_path = "{}/{}".format(
                                        temporary_dir_client,
                                        local_benchmark_output_filename,
                                    )
                                logging.info(
                                    "Reading results json from {}".format(
                                        full_result_path
                                    )
                                )

                                with open(
                                    full_result_path,
                                    "r",
                                ) as json_file:
                                    results_dict = json.load(json_file)
                                    print_results_table_stdout(
                                        benchmark_config,
                                        default_metrics,
                                        results_dict,
                                        setup_type,
                                        test_name,
                                        None,
                                    )

                                dataset_load_duration_seconds = 0
                                try:
                                    exporter_datasink_common(
                                        benchmark_config,
                                        benchmark_duration_seconds,
                                        build_variant_name,
                                        datapoint_time_ms,
                                        dataset_load_duration_seconds,
                                        datasink_conn,
                                        datasink_push_results_redistimeseries,
                                        git_branch,
                                        git_version,
                                        metadata,
                                        redis_conns,
                                        results_dict,
                                        running_platform,
                                        setup_name,
                                        setup_type,
                                        test_name,
                                        tf_github_org,
                                        tf_github_repo,
                                        tf_triggering_env,
                                        topology_spec_name,
                                        default_metrics,
                                        git_hash,
                                    )
                                    r.shutdown(save=False)

                                except redis.exceptions.ConnectionError as e:
                                    logging.critical(
                                        "Some unexpected exception was caught during metric fetching. Skipping it..."
                                    )
                                    logging.critical(
                                        f"Exception type: {type(e).__name__}"
                                    )
                                    logging.critical(f"Exception message: {str(e)}")
                                    logging.critical("Traceback details:")
                                    logging.critical(traceback.format_exc())
                                    print("-" * 60)
                                    traceback.print_exc(file=sys.stdout)
                                    print("-" * 60)

                                test_result = True
                                total_test_suite_runs = total_test_suite_runs + 1

                            except Exception as e:
                                logging.critical(
                                    "Some unexpected exception was caught during local work. Failing test...."
                                )
                                logging.critical(f"Exception type: {type(e).__name__}")
                                logging.critical(f"Exception message: {str(e)}")
                                logging.critical("Traceback details:")
                                logging.critical(traceback.format_exc())
                                print("-" * 60)
                                traceback.print_exc(file=sys.stdout)
                                print("-" * 60)
                                if redis_container is not None:
                                    logging.critical("Printing redis container log....")

                                    print("-" * 60)
                                    try:
                                        print(
                                            redis_container.logs(
                                                stdout=True, stderr=True
                                            )
                                        )
                                        redis_container.remove()
                                    except docker.errors.NotFound:
                                        logging.info(
                                            "When trying to fetch logs from DB container with id {} and image {} it was already stopped".format(
                                                redis_container.id,
                                                redis_container.image,
                                            )
                                        )
                                    pass

                                    print("-" * 60)

                                test_result = False
                            # tear-down
                            logging.info("Tearing down setup")
                            if docker_keep_env is False:
                                for redis_container in redis_containers:
                                    try:
                                        redis_container.stop()
                                        redis_container.remove()
                                    except docker.errors.NotFound:
                                        logging.info(
                                            "When trying to stop DB container with id {} and image {} it was already stopped".format(
                                                redis_container.id,
                                                redis_container.image,
                                            )
                                        )
                                        pass

                                for redis_container in client_containers:
                                    if type(redis_container) == Container:
                                        try:
                                            redis_container.stop()
                                            redis_container.remove()
                                        except docker.errors.NotFound:
                                            logging.info(
                                                "When trying to stop Client container with id {} and image {} it was already stopped".format(
                                                    redis_container.id,
                                                    redis_container.image,
                                                )
                                            )
                                            pass
                                logging.info(
                                    "Removing temporary dirs {} and {}".format(
                                        temporary_dir, temporary_dir_client
                                    )
                                )

                                shutil.rmtree(temporary_dir, ignore_errors=True)
                                shutil.rmtree(temporary_dir_client, ignore_errors=True)

                            overall_result &= test_result

                    github_event_conn.lrem(stream_test_list_running, 1, test_name)
                    github_event_conn.lpush(stream_test_list_completed, test_name)
                    github_event_conn.expire(
                        stream_test_list_completed, REDIS_BINS_EXPIRE_SECS
                    )
                    if test_result is False:
                        github_event_conn.lpush(stream_test_list_failed, test_name)
                        failed_tests = failed_tests + 1
                        logging.warning(
                            f"updating key {stream_test_list_failed} with the failed test: {test_name}. Total failed tests {failed_tests}."
                        )
                    pending_tests = pending_tests - 1

                    benchmark_suite_end_datetime = datetime.datetime.utcnow()
                    benchmark_suite_duration = (
                        benchmark_suite_end_datetime - benchmark_suite_start_datetime
                    )
                    benchmark_suite_duration_secs = (
                        benchmark_suite_duration.total_seconds()
                    )

                    # update on github if needed
                    if is_actionable_pr:
                        comment_body = generate_benchmark_started_pr_comment(
                            stream_id,
                            pending_tests,
                            len(filtered_test_files),
                            failed_tests,
                            benchmark_suite_start_datetime,
                            benchmark_suite_duration_secs,
                        )
                        update_comment_if_needed(
                            auto_approve_github,
                            comment_body,
                            old_benchmark_run_comment_body,
                            benchmark_run_comment,
                            verbose,
                        )
                        logging.info(
                            f"Updated github comment with latest test info {benchmark_run_comment.html_url}"
                        )

                        ###########################
                        # regression part
                        ###########################
                        fn = check_regression_comment
                        (
                            contains_regression_comment,
                            github_pr,
                            is_actionable_pr,
                            old_regression_comment_body,
                            pr_link,
                            regression_comment,
                        ) = check_github_available_and_actionable(
                            fn,
                            github_token,
                            pull_request,
                            tf_github_org,
                            tf_github_repo,
                            verbose,
                        )
                        logging.info("Preparing regression info for the data available")
                        print_improvements_only = False
                        print_regressions_only = False
                        skip_unstable = False
                        regressions_percent_lower_limit = 10.0
                        simplify_table = False
                        testname_regex = ""
                        test = ""
                        last_n_baseline = 1
                        last_n_comparison = 31
                        use_metric_context_path = False
                        baseline_tag = None
                        baseline_deployment_name = "oss-standalone"
                        comparison_deployment_name = "oss-standalone"
                        metric_name = "ALL_STATS.Totals.Ops/sec"
                        metric_mode = "higher-better"
                        to_date = datetime.datetime.utcnow()
                        from_date = to_date - datetime.timedelta(days=180)
                        baseline_branch = default_baseline_branch
                        comparison_tag = git_version
                        comparison_branch = git_branch
                        to_ts_ms = None
                        from_ts_ms = None

                        (
                            detected_regressions,
                            table_output,
                            improvement_list,
                            regressions_list,
                            total_stable,
                            total_unstable,
                            total_comparison_points,
                        ) = compute_regression_table(
                            datasink_conn,
                            tf_github_org,
                            tf_github_repo,
                            tf_triggering_env,
                            metric_name,
                            comparison_branch,
                            baseline_branch,
                            None,  # we only compare by branch on CI automation
                            None,  # we only compare by branch on CI automation
                            baseline_deployment_name,
                            comparison_deployment_name,
                            print_improvements_only,
                            print_regressions_only,
                            skip_unstable,
                            regressions_percent_lower_limit,
                            simplify_table,
                            test,
                            testname_regex,
                            verbose,
                            last_n_baseline,
                            last_n_comparison,
                            metric_mode,
                            from_date,
                            from_ts_ms,
                            to_date,
                            to_ts_ms,
                            use_metric_context_path,
                            running_platform,
                        )
                        total_regressions = len(regressions_list)
                        total_improvements = len(improvement_list)
                        auto_approve = True
                        grafana_link_base = "https://benchmarksredisio.grafana.net/d/1fWbtb7nz/experimental-oss-spec-benchmarks"
                        try:
                            prepare_regression_comment(
                                auto_approve,
                                baseline_branch,
                                baseline_tag,
                                comparison_branch,
                                comparison_tag,
                                contains_regression_comment,
                                github_pr,
                                grafana_link_base,
                                is_actionable_pr,
                                old_regression_comment_body,
                                pr_link,
                                regression_comment,
                                datasink_conn,
                                running_platform,
                                table_output,
                                tf_github_org,
                                tf_github_repo,
                                tf_triggering_env,
                                total_comparison_points,
                                total_improvements,
                                total_regressions,
                                total_stable,
                                total_unstable,
                                verbose,
                                regressions_percent_lower_limit,
                                regressions_list,
                            )
                        except Exception as e:
                            logging.error(
                                "Failed to produce regression comment but continuing... Error: {}".format(
                                    e.__str__()
                                )
                            )
                    logging.info(
                        f"Added test named {test_name} to the completed test list in key {stream_test_list_completed}"
                    )
        else:
            logging.error("Missing run image information within received message.")

    except Exception as e:
        logging.critical(
            "Some unexpected exception was caught "
            "during local work on stream {}. Failing test....".format(stream_id)
        )
        logging.critical(f"Exception type: {type(e).__name__}")
        logging.critical(f"Exception message: {str(e)}")
        logging.critical("Traceback details:")
        logging.critical(traceback.format_exc())
        print("-" * 60)
        traceback.print_exc(file=sys.stdout)
        print("-" * 60)
        overall_result = False
    return stream_id, overall_result, total_test_suite_runs


def start_redis_container(
    command_str,
    db_cpuset_cpus,
    docker_client,
    mnt_point,
    redis_containers,
    run_image,
    temporary_dir,
    auto_remove=False,
):
    logging.info(
        "Running redis-server on docker image {} (cpuset={}) with the following args: {}".format(
            run_image, db_cpuset_cpus, command_str
        )
    )
    volumes = {}
    working_dir = "/"
    if mnt_point != "":
        volumes = {
            temporary_dir: {
                "bind": mnt_point,
                "mode": "rw",
            },
        }
        logging.info(f"setting volume as follow: {volumes}. working_dir={mnt_point}")
        working_dir = mnt_point
    redis_container = docker_client.containers.run(
        image=run_image,
        volumes=volumes,
        auto_remove=auto_remove,
        privileged=True,
        working_dir=mnt_point,
        command=command_str,
        network_mode="host",
        detach=True,
        cpuset_cpus=db_cpuset_cpus,
        pid_mode="host",
        publish_all_ports=True,
    )
    time.sleep(5)
    redis_containers.append(redis_container)
    return redis_container


def filter_test_files(
    defaults_filename,
    priority_lower_limit,
    priority_upper_limit,
    tests_regexp,
    testsuite_spec_files,
    command_groups_regexp=None,
):
    filtered_test_files = []
    for test_file in testsuite_spec_files:
        if defaults_filename in test_file:
            continue

        if tests_regexp != ".*":
            logging.info(
                "Filtering all tests via a regular expression: {}".format(tests_regexp)
            )
            tags_regex_string = re.compile(tests_regexp)

            match_obj = re.search(tags_regex_string, test_file)
            if match_obj is None:
                logging.info(
                    "Skipping {} given it does not match regex {}".format(
                        test_file, tests_regexp
                    )
                )
                continue

        with open(test_file, "r") as stream:
            (
                result,
                benchmark_config,
                test_name,
            ) = get_final_benchmark_config(None, None, stream, "")
            if result is False:
                logging.error(
                    "Skipping {} given there were errors while calling get_final_benchmark_config()".format(
                        test_file
                    )
                )
                continue

            if command_groups_regexp is not None:
                logging.info(
                    "Filtering all test command groups via a regular expression: {}".format(
                        command_groups_regexp
                    )
                )
                if "tested-groups" in benchmark_config:
                    command_groups = benchmark_config["tested-groups"]
                    logging.info(
                        f"The file {test_file} (test name = {test_name}) contains the following groups: {command_groups}"
                    )
                    groups_regex_string = re.compile(command_groups_regexp)
                    found = False
                    for command_group in command_groups:
                        match_obj = re.search(groups_regex_string, command_group)
                        if match_obj is not None:
                            found = True
                            logging.info(f"found the command group {command_group}")
                    if found is False:
                        logging.info(
                            f"Skipping {test_file} given the following groups: {command_groups} does not match command group regex {command_groups_regexp}"
                        )
                        continue
                else:
                    logging.warning(
                        f"The file {test_file} (test name = {test_name}) does not contain the property 'tested-groups'. Cannot filter based uppon groups..."
                    )

            if "priority" in benchmark_config:
                priority = benchmark_config["priority"]

                if priority is not None:
                    if priority > priority_upper_limit:
                        logging.warning(
                            "Skipping test {} giving the priority limit ({}) is above the priority value ({})".format(
                                test_name, priority_upper_limit, priority
                            )
                        )

                        continue
                    if priority < priority_lower_limit:
                        logging.warning(
                            "Skipping test {} giving the priority limit ({}) is bellow the priority value ({})".format(
                                test_name, priority_lower_limit, priority
                            )
                        )

                        continue
                    logging.info(
                        "Test {} priority ({}) is within the priority limit [{},{}]".format(
                            test_name,
                            priority,
                            priority_lower_limit,
                            priority_upper_limit,
                        )
                    )
        filtered_test_files.append(test_file)
    return filtered_test_files


def data_prepopulation_step(
    benchmark_config,
    benchmark_tool_workdir,
    client_cpuset_cpus,
    docker_client,
    git_hash,
    port,
    temporary_dir,
    test_name,
):
    # setup the benchmark
    (
        start_time,
        start_time_ms,
        start_time_str,
    ) = get_start_time_vars()
    local_benchmark_output_filename = get_local_run_full_filename(
        start_time_str,
        git_hash,
        "preload__" + test_name,
        "oss-standalone",
    )
    preload_image = extract_client_container_image(
        benchmark_config["dbconfig"], "preload_tool"
    )
    preload_tool = extract_client_tool(benchmark_config["dbconfig"], "preload_tool")
    full_benchmark_path = "/usr/local/bin/{}".format(preload_tool)
    client_mnt_point = "/mnt/client/"
    if "memtier_benchmark" in preload_tool:
        (
            _,
            preload_command_str,
            _,
        ) = prepare_memtier_benchmark_parameters(
            benchmark_config["dbconfig"]["preload_tool"],
            full_benchmark_path,
            port,
            "localhost",
            None,
            local_benchmark_output_filename,
            False,
        )

        logging.info(
            "Using docker image {} as benchmark PRELOAD image (cpuset={}) with the following args: {}".format(
                preload_image,
                client_cpuset_cpus,
                preload_command_str,
            )
        )
        # run the benchmark
        preload_start_time = datetime.datetime.now()

        client_container_stdout = docker_client.containers.run(
            image=preload_image,
            volumes={
                temporary_dir: {
                    "bind": client_mnt_point,
                    "mode": "rw",
                },
            },
            auto_remove=True,
            privileged=True,
            working_dir=benchmark_tool_workdir,
            command=preload_command_str,
            network_mode="host",
            detach=False,
            cpuset_cpus=client_cpuset_cpus,
        )

        preload_end_time = datetime.datetime.now()
        preload_duration_seconds = calculate_client_tool_duration_and_check(
            preload_end_time, preload_start_time, "Preload", False
        )
        logging.info(
            "Tool {} seconds to load data. Output {}".format(
                preload_duration_seconds,
                client_container_stdout,
            )
        )


def get_benchmark_specs(testsuites_folder):
    files = pathlib.Path(testsuites_folder).glob("*.yml")
    files = [str(x) for x in files]
    logging.info(
        "Running all specified benchmarks: {}".format(" ".join([str(x) for x in files]))
    )
    return files
